# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from distutils.version import LooseVersion
import numpy as np
import pytest
import scipy

from hyperspy.learn.svd_pca import svd_pca
from hyperspy.misc.machine_learning.import_sklearn import sklearn_installed


class TestSVDPCA:
    def setup_method(self, method):
        # Define shape etc.
        m = 100  # Dimensionality
        n = 105  # Number of samples
        r = 3

        self.rng = np.random.RandomState(101)
        U = self.rng.randn(m, r)
        V = self.rng.randn(n, r)

        self.m = m
        self.n = n
        self.rank = r
        self.X = U @ V.T

        self.X_mean_0 = self.X.mean(axis=0)[np.newaxis, :]
        self.X_mean_1 = self.X.mean(axis=1)[:, np.newaxis]

        # Test tolerance
        self.tol = 1e-3 * (self.m * self.n)

    @pytest.mark.parametrize("output_dimension", [None, 3])
    @pytest.mark.parametrize("auto_transpose", [True, False])
    @pytest.mark.parametrize("centre", [None, "signal", "navigation"])
    @pytest.mark.parametrize("u_based_decision", [True, False])
    def test_full(self, output_dimension, auto_transpose, centre, u_based_decision):
        factors, loadings, explained_variance, mean = svd_pca(
            self.X,
            output_dimension=output_dimension,
            svd_solver="full",
            auto_transpose=auto_transpose,
            centre=centre,
            u_based_decision=u_based_decision,
        )
        X = loadings @ factors.T

        # Check the low-rank component MSE
        normX = np.linalg.norm(X - self.X)
        assert normX < self.tol

        # Check singular values
        explained_variance_norm = explained_variance / np.sum(explained_variance)
        np.testing.assert_allclose(explained_variance_norm[: self.rank].sum(), 1.0)

        if centre is None:
            assert mean is None
        elif centre == "features":
            np.testing.assert_allclose(mean, self.X_mean_1)
        elif centre == "samples":
            np.testing.assert_allclose(mean, self.X_mean_0)

    @pytest.mark.xfail(LooseVersion(scipy.__version__) < LooseVersion("1.4.0"),
                       raises=ValueError,
                       reason="This test requires Scipy >= 1.4.0")
    @pytest.mark.parametrize("output_dimension", [None, 3])
    @pytest.mark.parametrize("auto_transpose", [True, False])
    @pytest.mark.parametrize("centre", [None, "signal", "navigation"])
    @pytest.mark.parametrize("u_based_decision", [True, False])
    def test_arpack(self, output_dimension, auto_transpose, centre, u_based_decision):
        factors, loadings, explained_variance, mean = svd_pca(
            self.X,
            output_dimension=output_dimension,
            svd_solver="arpack",
            auto_transpose=auto_transpose,
            centre=centre,
            u_based_decision=u_based_decision,
        )
        X = loadings @ factors.T

        # Check the low-rank component MSE
        normX = np.linalg.norm(X - self.X)
        assert normX < self.tol

        # Check singular values
        explained_variance_norm = explained_variance / np.sum(explained_variance)
        np.testing.assert_allclose(explained_variance_norm[: self.rank].sum(), 1.0)

    @pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
    @pytest.mark.parametrize("output_dimension", [None, 3])
    @pytest.mark.parametrize("auto_transpose", [True, False])
    @pytest.mark.parametrize("centre", [None, "signal", "navigation"])
    def test_randomized(self, output_dimension, auto_transpose, centre):
        factors, loadings, explained_variance, mean = svd_pca(
            self.X,
            output_dimension=output_dimension,
            svd_solver="randomized",
            auto_transpose=auto_transpose,
            centre=centre,
        )
        X = loadings @ factors.T

        # Check the low-rank component MSE
        normX = np.linalg.norm(X - self.X)
        assert normX < self.tol

        # Check singular values
        explained_variance_norm = explained_variance / np.sum(explained_variance)
        np.testing.assert_allclose(explained_variance_norm[: self.rank].sum(), 1.0)

    @pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
    def test_solver_auto(self):
        # Uses "full"
        U = self.rng.randn(100, 5)
        V = self.rng.randn(100, 5)
        X = U @ V.T
        factors, loadings, _, _ = svd_pca(X, output_dimension=5, svd_solver="auto")
        Y = loadings @ factors.T
        normX = np.linalg.norm(X - Y)
        assert normX < self.tol

        # Uses "randomized"
        U = self.rng.randn(501, 5)
        V = self.rng.randn(100, 5)
        X = U @ V.T
        factors, loadings, _, _ = svd_pca(X, output_dimension=5, svd_solver="auto")
        Y = loadings @ factors.T
        normX = np.linalg.norm(X - Y)
        assert normX < self.tol

        # Uses full
        U = self.rng.randn(501, 5)
        V = self.rng.randn(100, 5)
        X = U @ V.T
        factors, loadings, _, _ = svd_pca(X, output_dimension=81, svd_solver="auto")
        Y = loadings @ factors.T
        normX = np.linalg.norm(X - Y)
        assert normX < self.tol

    def test_arpack_error(self):
        pytest.importorskip("scipy", minversion="1.4.0")
        with pytest.raises(
            ValueError, match="requires output_dimension to be strictly"
        ):
            _ = svd_pca(
                self.X, output_dimension=min(self.X.shape) + 1, svd_solver="arpack"
            )

    def test_centre_error(self):
        with pytest.raises(ValueError, match="'centre' must be one of"):
            _ = svd_pca(self.X, centre="random")



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

"""Common docstring snippets for model.

"""

FIT_PARAMETERS_ARG = """optimizer : str or None, default None
            The optimization algorithm used to perform the fitting.

                * "lm" performs least-squares optimization using the
                  Levenberg-Marquardt algorithm, and supports bounds
                  on parameters.
                * "trf" performs least-squares optimization using the
                  Trust Region Reflective algorithm, and supports
                  bounds on parameters.
                * "dogbox" performs least-squares optimization using the
                  dogleg algorithm with rectangular trust regions, and
                  supports bounds on parameters.
                * "odr" performs the optimization using the orthogonal
                  distance regression (ODR) algorithm. It does not support
                  bounds on parameters. See :py:mod:`scipy.odr` for more details.
                * All of the available methods for :py:func:`scipy.optimize.minimize`
                  can be used here. See the :ref:`User Guide <model.fitting>`
                  documentation for more details.
                * "Differential Evolution" is a global optimization method.
                  It does support bounds on parameters. See
                  :py:func:`scipy.optimize.differential_evolution` for more
                  details on available options.
                * "Dual Annealing" is a global optimization method.
                  It does support bounds on parameters. See
                  :py:func:`scipy.optimize.dual_annealing` for more
                  details on available options. Requires ``scipy >= 1.2.0``.
                * "SHGO" (simplicial homology global optimization" is a global
                  optimization method. It does support bounds on parameters. See
                  :py:func:`scipy.optimize.shgo` for more details on available
                  options. Requires ``scipy >= 1.2.0``.

        loss_function : {"ls", "ML-poisson", "huber", callable}, default "ls"
            The loss function to use for minimization. Only ``"ls"`` is available
            if ``optimizer`` is one of ``["lm", "trf", "dogbox", "odr"]``.

                * "ls" minimizes the least-squares loss function.
                * "ML-poisson" minimizes the negative log-likelihood for
                  Poisson-distributed data. Also known as Poisson maximum
                  likelihood estimation (MLE).
                * "huber" minimize the Huber loss function. The delta value
                  of the Huber function is controlled by the ``huber_delta``
                  keyword argument (the default value is 1.0).
                * callable supports passing your own minimization function.

        grad : {"fd", "analytical", callable, None}, default "fd"
            Whether to use information about the gradient of the loss function
            as part of the optimization. This parameter has no effect if
            ``optimizer`` is a derivative-free or global optimization method.

                * "fd" uses a finite difference scheme (if available) for numerical
                  estimation of the gradient. The scheme can be further controlled
                  with the ``fd_scheme`` keyword argument.
                * "analytical" uses the analytical gradient (if available) to speed
                  up the optimization, since the gradient does not need to be estimated.
                * callable should be a function that returns the gradient vector.
                * None means that no gradient information is used or estimated. Not
                  available if ``optimizer`` is in ``["lm", "trf", ``"dogbox"``].

        bounded : bool, default False
            If True, performs bounded parameter optimization if
            supported by ``optimizer``.
        update_plot : bool, default False
            If True, the plot is updated during the optimization
            process. It slows down the optimization, but it enables
            visualization of the optimization progress.
        print_info : bool, default False
            If True, print information about the fitting results, which
            are also stored in ``model.fit_output`` in the form of
            a :py:class:`scipy.optimize.OptimizeResult` object.
        return_info : bool, default True
            If True, returns the fitting results in the form of
            a :py:class:`scipy.optimize.OptimizeResult` object.
        fd_scheme : str {"2-point", "3-point", "cs"}, default "2-point"
            If ``grad='fd'``, selects the finite difference scheme to use.
            See :py:func:`scipy.optimize.minimize` for details. Ignored if
            ``optimizer`` is ``"lm"``, ``"trf"`` or ``"dogbox"``.
        **kwargs : keyword arguments
            Any extra keyword argument will be passed to the chosen
            optimizer. For more information, read the docstring of the
            optimizer of your choice in :py:mod:`scipy.optimize`."""



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import math
import sympy

from hyperspy._components.expression import Expression
from hyperspy._components.gaussian import _estimate_gaussian_parameters
from distutils.version import LooseVersion

sqrt2pi = math.sqrt(2 * math.pi)
sigma2fwhm = 2 * math.sqrt(2 * math.log(2))


class Voigt(Expression):

    r"""Voigt component.

    Symmetric peak shape based on the convolution of a Lorentzian and Normal
    (Gaussian) distribution:

    .. math::
        f(x) = G(x) \cdot L(x)

    where :math:`G(x)` is the Gaussian function and :math:`L(x)` is the
    Lorentzian function. In this case using an approximate formula by David
    (see Notes). This approximation improves on the pseudo-Voigt function
    (linear combination instead of convolution of the distributions) and is,
    to a very good approximation, equivalent to a Voigt function:

    .. math::
        z(x) &= \frac{x + i \gamma}{\sqrt{2} \sigma} \\
        w(z) &= \frac{e^{-z^2} \text{erfc}(-i z)}{\sqrt{2 \pi} \sigma} \\
        f(x) &= A \cdot \Re\left\{ w \left[ z(x - x_0) \right] \right\}


    ============== =============
    Variable        Parameter
    ============== =============
    :math:`x_0`     centre
    :math:`A`       area
    :math:`\gamma`  gamma
    :math:`\sigma`  sigma
    ============== =============


    Parameters
    -----------
    centre : float
        Location of the maximum of the peak.
    area : float
        Intensity below the peak.
    gamma : float
        :math:`\gamma` = HWHM of the Lorentzian distribution.
    sigma: float
        :math:`2 \sigma \sqrt{(2 \log(2))}` = FWHM of the Gaussian distribution.


    For convenience the `gwidth` and `lwidth` attributes can also be used to
    set and get the FWHM of the Gaussian and Lorentzian parts of the
    distribution, respectively. For backwards compatability, `FWHM` is another
    alias for the Gaussian width.

    Notes
    -----
    W.I.F. David, J. Appl. Cryst. (1986). 19, 63-64,
    doi:10.1107/S0021889886089999
    """

    def __init__(self, centre=10., area=1., gamma=0.2, sigma=0.1,
                 module=["numpy", "scipy"], **kwargs):
        # Not to break scripts once we remove the legacy Voigt
        if "legacy" in kwargs:
            del kwargs["legacy"]
        if LooseVersion(sympy.__version__) < LooseVersion("1.3"):
            raise ImportError("The `Voigt` component requires "
                              "SymPy >= 1.3")
        # We use `_gamma` internally to workaround the use of the `gamma`
        # function in sympy
        super(Voigt, self).__init__(
            expression="area * real(V); \
                V = wofz(z) / (sqrt(2.0 * pi) * sigma); \
                z = (x - centre + 1j * _gamma) / (sigma * sqrt(2.0))",
            name="Voigt",
            centre=centre,
            area=area,
            gamma=gamma,
            sigma=sigma,
            position="centre",
            module=module,
            autodoc=False,
            rename_pars={"_gamma": "gamma"},
            **kwargs,
        )

        # Boundaries
        self.area.bmin = 0.
        self.gamma.bmin = 0.
        self.sigma.bmin = 0.

        self.isbackground = False
        self.convolved = True

    def estimate_parameters(self, signal, x1, x2, only_current=False):
        """Estimate the Voigt function by calculating the momenta of the
        Gaussian.

        Parameters
        ----------
        signal : Signal1D instance
        x1 : float
            Defines the left limit of the spectral range to use for the
            estimation.
        x2 : float
            Defines the right limit of the spectral range to use for the
            estimation.

        only_current : bool
            If False estimates the parameters for the full dataset.

        Returns
        -------
         : bool
            Exit status required for the :meth:`remove_background` function.

        Notes
        -----
        Adapted from http://www.scipy.org/Cookbook/FittingData

        Examples
        --------

        >>> g = hs.model.components1D.Voigt(legacy=False)
        >>> x = np.arange(-10, 10, 0.01)
        >>> data = np.zeros((32, 32, 2000))
        >>> data[:] = g.function(x).reshape((1, 1, 2000))
        >>> s = hs.signals.Signal1D(data)
        >>> s.axes_manager[-1].offset = -10
        >>> s.axes_manager[-1].scale = 0.01
        >>> g.estimate_parameters(s, -10, 10, False)

        """
        super(Voigt, self)._estimate_parameters(signal)
        axis = signal.axes_manager.signal_axes[0]
        centre, height, sigma = _estimate_gaussian_parameters(signal, x1, x2,
                                                              only_current)

        if only_current is True:
            self.centre.value = centre
            self.sigma.value = sigma
            self.area.value = height * sigma * sqrt2pi
            if self.binned:
                self.area.value /= axis.scale
            return True
        else:
            if self.area.map is None:
                self._create_arrays()
            self.area.map['values'][:] = height * sigma * sqrt2pi
            if self.binned:
                self.area.map['values'][:] /= axis.scale
            self.area.map['is_set'][:] = True
            self.sigma.map['values'][:] = sigma
            self.sigma.map['is_set'][:] = True
            self.centre.map['values'][:] = centre
            self.centre.map['is_set'][:] = True
            self.fetch_stored_values()
            return True

    @property
    def gwidth(self):
        return self.sigma.value * sigma2fwhm

    @gwidth.setter
    def gwidth(self, value):
        self.sigma.value = value / sigma2fwhm

    @property
    def FWHM(self):
        return self.sigma.value * sigma2fwhm

    @FWHM.setter
    def FWHM(self, value):
        self.sigma.value = value / sigma2fwhm

    @property
    def lwidth(self):
        return self.gamma.value * 2

    @lwidth.setter
    def lwidth(self, value):
        self.gamma.value = value / 2



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging
from unittest import mock

import numpy as np
import pytest

from hyperspy.decorators import lazifyTestClass
from hyperspy.signals import Signal1D, Signal2D


def _verify_test_sum_x_E(self, s):
    np.testing.assert_array_equal(self.signal.data.sum(), s.data)
    assert s.data.ndim == 1
    # Check that there is still one signal axis.
    assert s.axes_manager.signal_dimension == 1


@lazifyTestClass
class Test2D:
    def setup_method(self, method):
        self.signal = Signal1D(np.arange(5 * 10).reshape(5, 10))  # dtype int
        self.signal.axes_manager[0].name = "x"
        self.signal.axes_manager[1].name = "E"
        self.signal.axes_manager[0].scale = 0.5
        self.data = self.signal.data.copy()

    def test_sum_x(self):
        s = self.signal.sum("x")
        np.testing.assert_array_equal(self.signal.data.sum(0), s.data)
        assert s.data.ndim == 1
        assert s.axes_manager.navigation_dimension == 0

    def test_sum_x_E(self):
        s = self.signal.sum(("x", "E"))
        _verify_test_sum_x_E(self, s)
        s = self.signal.sum((0, "E"))
        _verify_test_sum_x_E(self, s)
        s = self.signal.sum((self.signal.axes_manager[0], "E"))
        _verify_test_sum_x_E(self, s)
        s = self.signal.sum("x").sum("E")
        _verify_test_sum_x_E(self, s)

    def test_axis_by_str(self):
        m = mock.Mock()
        s1 = self.signal.deepcopy()
        s1.events.data_changed.connect(m.data_changed)
        s2 = self.signal.deepcopy()
        s1.crop(0, 2, 4)
        assert m.data_changed.called
        s2.crop("x", 2, 4)
        np.testing.assert_array_almost_equal(s1.data, s2.data)

    def test_crop_int(self):
        s = self.signal
        d = self.data
        s.crop(0, 2, 4)
        np.testing.assert_array_almost_equal(s.data, d[2:4, :])

    def test_crop_float(self):
        s = self.signal
        d = self.data
        s.crop(0, 2, 2.0)
        np.testing.assert_array_almost_equal(s.data, d[2:4, :])

    def test_crop_start_end_equal(self):
        s = self.signal
        with pytest.raises(ValueError):
            s.crop(0, 2, 2)
        with pytest.raises(ValueError):
            s.crop(0, 2.0, 2.0)

    def test_crop_float_no_unit_convertion_signal1D(self):
        # Should convert the unit to eV
        d = np.arange(5 * 10 * 2000).reshape(5, 10, 2000)
        s = Signal1D(d)
        s.axes_manager.signal_axes[0].name = "E"
        s.axes_manager.signal_axes[0].scale = 0.05
        s.axes_manager.signal_axes[0].units = "keV"
        s.crop("E", 0.0, 1.0, convert_units=False)
        np.testing.assert_almost_equal(s.axes_manager.signal_axes[0].scale, 0.05)
        assert s.axes_manager.signal_axes[0].units == "keV"
        np.testing.assert_allclose(s.data, d[:, :, :20])

        # Should keep the unit to keV
        s = Signal1D(d)
        s.axes_manager.signal_axes[0].name = "E"
        s.axes_manager.signal_axes[0].scale = 0.05
        s.axes_manager.signal_axes[0].units = "keV"
        s.crop("E", 0.0, 50.0, convert_units=False)
        np.testing.assert_almost_equal(s.axes_manager.signal_axes[0].scale, 0.05)
        assert s.axes_manager.signal_axes[0].units == "keV"
        np.testing.assert_allclose(s.data, d[:, :, :1000])

    def test_crop_float_unit_convertion_signal1D(self):
        # Should convert the unit to eV
        d = np.arange(5 * 10 * 2000).reshape(5, 10, 2000)
        s = Signal1D(d)
        s.axes_manager.signal_axes[0].name = "E"
        s.axes_manager.signal_axes[0].scale = 0.05
        s.axes_manager.signal_axes[0].units = "keV"
        s.crop("E", 0.0, 1.0, convert_units=True)
        np.testing.assert_almost_equal(s.axes_manager.signal_axes[0].scale, 50.0)
        assert s.axes_manager.signal_axes[0].units == "eV"
        np.testing.assert_allclose(s.data, d[:, :, :20])

        # Should keep the unit to keV
        s = Signal1D(d)
        s.axes_manager.signal_axes[0].name = "E"
        s.axes_manager.signal_axes[0].scale = 0.05
        s.axes_manager.signal_axes[0].units = "keV"
        s.crop("E", 0.0, 50.0, convert_units=True)
        np.testing.assert_almost_equal(s.axes_manager.signal_axes[0].scale, 0.05)
        assert s.axes_manager.signal_axes[0].units == "keV"
        np.testing.assert_allclose(s.data, d[:, :, :1000])

    def test_crop_float_no_unit_convertion_signal2D(self):
        # Should convert the unit to nm
        d = np.arange(512 * 512).reshape(512, 512)
        s = Signal2D(d)
        s.axes_manager[0].name = "x"
        s.axes_manager[0].scale = 0.01
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].name = "y"
        s.axes_manager[1].scale = 0.01
        s.axes_manager[1].units = "µm"
        s.crop(0, 0.0, 0.5, convert_units=False)
        s.crop(1, 0.0, 0.5, convert_units=False)
        np.testing.assert_almost_equal(s.axes_manager[0].scale, 0.01)
        assert s.axes_manager[0].units == "µm"
        np.testing.assert_allclose(s.data, d[:50, :50])

        # Should keep the unit to µm
        d = np.arange(512 * 512).reshape(512, 512)
        s = Signal2D(d)
        s.axes_manager[0].name = "x"
        s.axes_manager[0].scale = 0.01
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].name = "y"
        s.axes_manager[1].scale = 0.01
        s.axes_manager[1].units = "µm"
        s.crop(0, 0.0, 5.0, convert_units=False)
        s.crop(1, 0.0, 5.0, convert_units=False)
        np.testing.assert_almost_equal(s.axes_manager[0].scale, 0.01)
        assert s.axes_manager[0].units == "µm"
        np.testing.assert_allclose(s.data, d[:500, :500])

    def test_crop_float_unit_convertion_signal2D(self):
        # Should convert the unit to nm
        d = np.arange(512 * 512).reshape(512, 512)
        s = Signal2D(d)
        s.axes_manager[0].name = "x"
        s.axes_manager[0].scale = 0.01
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].name = "y"
        s.axes_manager[1].scale = 0.01
        s.axes_manager[1].units = "µm"
        s.crop(0, 0.0, 0.5, convert_units=True)  # also convert the other axis
        s.crop(1, 0.0, 500.0, convert_units=True)
        np.testing.assert_almost_equal(s.axes_manager[0].scale, 10.0)
        np.testing.assert_almost_equal(s.axes_manager[1].scale, 10.0)
        assert s.axes_manager[0].units == "nm"
        assert s.axes_manager[1].units == "nm"
        np.testing.assert_allclose(s.data, d[:50, :50])

        # Should keep the unit to µm
        d = np.arange(512 * 512).reshape(512, 512)
        s = Signal2D(d)
        s.axes_manager[0].name = "x"
        s.axes_manager[0].scale = 0.01
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].name = "y"
        s.axes_manager[1].scale = 0.01
        s.axes_manager[1].units = "µm"
        s.crop(0, 0.0, 5.0, convert_units=True)
        s.crop(1, 0.0, 5.0, convert_units=True)
        np.testing.assert_almost_equal(s.axes_manager[0].scale, 0.01)
        np.testing.assert_almost_equal(s.axes_manager[1].scale, 0.01)
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.data, d[:500, :500])

    def test_crop_image_unit_convertion_signal2D(self):
        # Should not convert the unit
        d = np.arange(512 * 512).reshape(512, 512)
        s = Signal2D(d)
        s.axes_manager[0].name = "x"
        s.axes_manager[0].scale = 0.01
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].name = "y"
        s.axes_manager[1].scale = 0.01
        s.axes_manager[1].units = "µm"
        s.crop_image(0, 0.5, 0.0, 0.5)
        np.testing.assert_almost_equal(s.axes_manager[0].scale, 0.01)
        np.testing.assert_almost_equal(s.axes_manager[1].scale, 0.01)
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.data, d[:50, :50])

        # Should convert the unit to nm
        d = np.arange(512 * 512).reshape(512, 512)
        s = Signal2D(d)
        s.axes_manager[0].name = "x"
        s.axes_manager[0].scale = 0.01
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].name = "y"
        s.axes_manager[1].scale = 0.01
        s.axes_manager[1].units = "µm"
        s.crop_image(0, 0.5, 0.0, 0.5, convert_units=True)
        np.testing.assert_almost_equal(s.axes_manager[0].scale, 10.0)
        np.testing.assert_almost_equal(s.axes_manager[1].scale, 10.0)
        assert s.axes_manager[0].units == "nm"
        assert s.axes_manager[1].units == "nm"
        np.testing.assert_allclose(s.data, d[:50, :50])

        # Should keep the unit to µm
        d = np.arange(512 * 512).reshape(512, 512)
        s = Signal2D(d)
        s.axes_manager[0].name = "x"
        s.axes_manager[0].scale = 0.01
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].name = "y"
        s.axes_manager[1].scale = 0.01
        s.axes_manager[1].units = "µm"
        s.crop_image(0, 5.0, 0.0, 5.0, convert_units=True)
        np.testing.assert_almost_equal(s.axes_manager[0].scale, 0.01)
        np.testing.assert_almost_equal(s.axes_manager[1].scale, 0.01)
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.data, d[:500, :500])

    def test_split_axis0(self):
        result = self.signal.split(0, 2)
        assert len(result) == 2
        np.testing.assert_array_almost_equal(result[0].data, self.data[:2, :])
        np.testing.assert_array_almost_equal(result[1].data, self.data[2:4, :])

    def test_split_axis1(self):
        result = self.signal.split(1, 2)
        assert len(result) == 2
        np.testing.assert_array_almost_equal(result[0].data, self.data[:, :5])
        np.testing.assert_array_almost_equal(result[1].data, self.data[:, 5:])

    def test_split_axisE(self):
        result = self.signal.split("E", 2)
        assert len(result) == 2
        np.testing.assert_array_almost_equal(result[0].data, self.data[:, :5])
        np.testing.assert_array_almost_equal(result[1].data, self.data[:, 5:])

    def test_split_default(self):
        result = self.signal.split()
        assert len(result) == 5
        np.testing.assert_array_almost_equal(result[0].data, self.data[0])

    def test_split_step_size_list(self):
        result = self.signal.split(step_sizes=[1, 2])
        assert len(result) == 2
        np.testing.assert_array_almost_equal(result[0].data, self.data[:1, :10])
        np.testing.assert_array_almost_equal(result[1].data, self.data[1:3, :10])

    def test_split_error(self):
        with pytest.raises(
            ValueError,
            match="You can define step_sizes or number_of_parts but not both",
        ):
            _ = self.signal.split(number_of_parts=2, step_sizes=2)

        with pytest.raises(
            ValueError, match="The number of parts is greater than the axis size.",
        ):
            _ = self.signal.split(number_of_parts=1e9)

    def test_histogram(self):
        result = self.signal.get_histogram(3)
        assert isinstance(result, Signal1D)
        np.testing.assert_array_equal(result.data, np.array([17, 16, 17]))
        assert result.metadata.Signal.binned

    def test_noise_variance_helpers(self):
        assert self.signal.get_noise_variance() is None
        self.signal.set_noise_variance(2)
        assert self.signal.get_noise_variance() == 2
        self.signal.set_noise_variance(self.signal)
        variance = self.signal.get_noise_variance()
        np.testing.assert_array_equal(variance.data, self.signal.data)
        self.signal.set_noise_variance(None)
        assert self.signal.get_noise_variance() is None

        with pytest.raises(ValueError, match="`variance` must be one of"):
            self.signal.set_noise_variance(np.array([0, 1, 2]))

    def test_estimate_poissonian_noise_copy_data(self):
        self.signal.estimate_poissonian_noise_variance()
        variance = self.signal.metadata.Signal.Noise_properties.variance
        assert variance.data is not self.signal.data

    def test_estimate_poissonian_noise_copy_data_helper_function(self):
        self.signal.estimate_poissonian_noise_variance()
        variance = self.signal.get_noise_variance()
        assert variance.data is not self.signal.data

    def test_estimate_poissonian_noise_noarg(self):
        self.signal.estimate_poissonian_noise_variance()
        variance = self.signal.metadata.Signal.Noise_properties.variance
        np.testing.assert_array_equal(variance.data, self.signal.data)

    def test_estimate_poissonian_noise_with_args(self):
        self.signal.estimate_poissonian_noise_variance(
            expected_value=self.signal,
            gain_factor=2,
            gain_offset=1,
            correlation_factor=0.5,
        )
        variance = self.signal.metadata.Signal.Noise_properties.variance
        np.testing.assert_array_equal(variance.data, (self.signal.data * 2 + 1) * 0.5)

    def test_unfold_image(self):
        s = self.signal
        if s._lazy:
            pytest.skip("LazySignals do not support folding")
        s = s.transpose(signal_axes=2)
        s.unfold()
        assert s.data.shape == (50,)

    def test_unfold_image_returns_true(self):
        s = self.signal
        if s._lazy:
            pytest.skip("LazySignals do not support folding")
        s = s.transpose(signal_axes=2)
        assert s.unfold()

    def test_print_summary(self):
        # Just test if it doesn't raise an exception
        self.signal._print_summary()

    def test_print_summary_statistics(self):
        # Just test if it doesn't raise an exception
        self.signal.print_summary_statistics()
        if self.signal._lazy:
            self.signal.print_summary_statistics(rechunk=False)

    def test_numpy_unfunc_one_arg_titled(self):
        self.signal.metadata.General.title = "yes"
        result = np.exp(self.signal)
        assert isinstance(result, Signal1D)
        np.testing.assert_array_equal(result.data, np.exp(self.signal.data))
        assert result.metadata.General.title == "exp(yes)"

    def test_numpy_unfunc_one_arg_untitled(self):
        result = np.exp(self.signal)
        assert result.metadata.General.title == "exp(Untitled Signal 1)"

    def test_numpy_unfunc_two_arg_titled(self):
        s1, s2 = self.signal.deepcopy(), self.signal.deepcopy()
        s1.metadata.General.title = "A"
        s2.metadata.General.title = "B"
        result = np.add(s1, s2)
        assert isinstance(result, Signal1D)
        np.testing.assert_array_equal(result.data, np.add(s1.data, s2.data))
        assert result.metadata.General.title == "add(A, B)"

    def test_numpy_unfunc_two_arg_untitled(self):
        s1, s2 = self.signal.deepcopy(), self.signal.deepcopy()
        result = np.add(s1, s2)
        assert (
            result.metadata.General.title == "add(Untitled Signal 1, Untitled Signal 2)"
        )

    def test_numpy_func(self):
        result = np.angle(self.signal)
        assert isinstance(result, np.ndarray)
        np.testing.assert_array_equal(result, np.angle(self.signal.data))

    def test_add_gaussian_noise(self):
        s = self.signal
        s.change_dtype("float64")
        kwargs = {}
        if s._lazy:
            data = s.data.compute()
            from dask.array.random import seed, normal

            kwargs["chunks"] = s.data.chunks
        else:
            data = s.data.copy()
            from numpy.random import seed, normal
        seed(1)
        s.add_gaussian_noise(std=1.0)
        seed(1)
        if s._lazy:
            s.compute()
        np.testing.assert_array_almost_equal(
            s.data - data, normal(scale=1.0, size=data.shape, **kwargs)
        )

    def test_add_gaussian_noise_seed(self):
        s = self.signal
        s.change_dtype("float64")
        kwargs = {}
        if s._lazy:
            data = s.data.compute()
            from dask.array.random import RandomState

            kwargs["chunks"] = s.data.chunks
            rng1 = RandomState(123)
            rng2 = RandomState(123)
        else:
            data = s.data.copy()
            rng1 = np.random.RandomState(123)
            rng2 = np.random.RandomState(123)

        s.add_gaussian_noise(std=1.0, random_state=rng1)
        if s._lazy:
            s.compute()

        np.testing.assert_array_almost_equal(
            s.data - data, rng2.normal(scale=1.0, size=data.shape, **kwargs)
        )

    def test_gaussian_noise_error(self):
        s = self.signal
        s.change_dtype("int64")
        with pytest.raises(TypeError, match="float datatype"):
            s.add_gaussian_noise(std=1.0)

    def test_add_poisson_noise(self):
        s = self.signal
        kwargs = {}
        if s._lazy:
            data = s.data.compute()
            from dask.array.random import seed, poisson

            kwargs["chunks"] = s.data.chunks
        else:
            data = s.data.copy()
            from numpy.random import seed, poisson
        seed(1)

        s.add_poissonian_noise(keep_dtype=False)

        if s._lazy:
            s.compute()
        seed(1)
        np.testing.assert_array_almost_equal(s.data, poisson(lam=data, **kwargs))
        s.change_dtype("float64")
        seed(1)

        s.add_poissonian_noise(keep_dtype=True)
        if s._lazy:
            s.compute()
        assert s.data.dtype == np.dtype("float64")

    def test_add_poisson_noise_seed(self):
        s = self.signal
        kwargs = {}
        if s._lazy:
            data = s.data.compute()
            from dask.array.random import RandomState

            kwargs["chunks"] = s.data.chunks
            rng1 = RandomState(123)
            rng2 = RandomState(123)
        else:
            data = s.data.copy()
            rng1 = np.random.RandomState(123)
            rng2 = np.random.RandomState(123)

        s.add_poissonian_noise(keep_dtype=False, random_state=rng1)

        if s._lazy:
            s.compute()

        np.testing.assert_array_almost_equal(s.data, rng2.poisson(lam=data, **kwargs))
        s.change_dtype("float64")

        s.add_poissonian_noise(keep_dtype=True, random_state=rng1)
        if s._lazy:
            s.compute()

        assert s.data.dtype == np.dtype("float64")

    def test_add_poisson_noise_warning(self, caplog):
        s = self.signal
        s.change_dtype("float64")

        with caplog.at_level(logging.WARNING):
            s.add_poissonian_noise(keep_dtype=True)

        assert "Changing data type from" in caplog.text

        with caplog.at_level(logging.WARNING):
            s.add_poissonian_noise(keep_dtype=False)

        assert "The data type changed from" in caplog.text



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import pytest

from hyperspy._signals.signal1d import Signal1D
from hyperspy._signals.signal2d import Signal2D
from hyperspy.components1d import Gaussian
from hyperspy.models.model1d import ComponentFit
from hyperspy.exceptions import SignalDimensionError


class TestFitOneComponent:

    def setup_method(self, method):
        g = Gaussian()
        g.A.value = 10000.0
        g.centre.value = 5000.0
        g.sigma.value = 500.0
        axis = np.arange(10000)
        s = Signal1D(g.function(axis))
        m = s.create_model()
        self.model = m
        self.g = g
        self.axis = axis

    @pytest.mark.parametrize("signal_range", [(4000, 6000), 'interactive'])
    def test_fit_component(self, signal_range):
        m = self.model
        axis = self.axis
        g = self.g

        g1 = Gaussian()
        m.append(g1)
        cf = ComponentFit(m, g1, signal_range=signal_range)
        if signal_range == 'interactive':
            cf.ss_left_value, cf.ss_right_value = (4000, 6000)
        cf._fit_fired()
        np.testing.assert_allclose(g.function(axis),
                                   g1.function(axis),
                                   rtol=0.0,
                                   atol=10e-3)

    def test_component_not_in_model(self):
        with pytest.raises(ValueError):
            self.model.fit_component(self.g)


def test_Component_fit_wrong_signal():
    s = Signal2D(np.arange(2*3*4).reshape(2, 3, 4))
    m = s.create_model()
    with pytest.raises(SignalDimensionError):
        ComponentFit(m, Gaussian())


class TestFitSeveralComponent:

    def setup_method(self, method):
        gs1 = Gaussian()
        gs1.A.value = 10000.0
        gs1.centre.value = 5000.0
        gs1.sigma.value = 500.0

        gs2 = Gaussian()
        gs2.A.value = 60000.0
        gs2.centre.value = 2000.0
        gs2.sigma.value = 300.0

        gs3 = Gaussian()
        gs3.A.value = 20000.0
        gs3.centre.value = 6000.0
        gs3.sigma.value = 100.0

        axis = np.arange(10000)
        total_signal = (gs1.function(axis) +
                        gs2.function(axis) +
                        gs3.function(axis))

        s = Signal1D(total_signal)
        m = s.create_model()

        g1 = Gaussian()
        g2 = Gaussian()
        g3 = Gaussian()
        m.append(g1)
        m.append(g2)
        m.append(g3)

        self.model = m
        self.gs1 = gs1
        self.gs2 = gs2
        self.gs3 = gs3
        self.g1 = g1
        self.g2 = g2
        self.g3 = g3
        self.axis = axis
        self.rtol = 0.01

    def test_fit_component_active_state(self):
        m = self.model
        axis = self.axis
        g1 = self.g1
        g2 = self.g2
        g3 = self.g3
        g2.active = True
        g3.active = False
        m.fit_component(g1, signal_range=(4500, 5200), fit_independent=True)
        np.testing.assert_allclose(self.gs1.function(axis),
                                   g1.function(axis),
                                   rtol=self.rtol,
                                   atol=10e-3)
        assert g1.active
        assert g2.active
        assert not g3.active

    def test_fit_component_free_state(self):
        m = self.model
        axis = self.axis
        g1 = self.g1
        g2 = self.g2
        g3 = self.g3
        g2.A.free = False
        g2.sigma.free = False
        m.fit_component(g1, signal_range=(4500, 5200))
        np.testing.assert_allclose(self.gs1.function(axis),
                                   g1.function(axis),
                                   rtol=self.rtol,
                                   atol=10e-3)

        assert g1.A.free
        assert g1.sigma.free
        assert g1.centre.free

        assert not g2.A.free
        assert not g2.sigma.free
        assert g2.centre.free

        assert g3.A.free
        assert g3.sigma.free
        assert g3.centre.free

    def test_fit_multiple_component(self):
        m = self.model
        g1 = self.g1
        g2 = self.g2
        g3 = self.g3
        m.fit_component(g1, signal_range=(4500, 5200))
        m.fit_component(g2, signal_range=(1500, 2200))
        m.fit_component(g3, signal_range=(5800, 6150))
        np.testing.assert_allclose(self.model.signal.data,
                                   m(),
                                   rtol=self.rtol,
                                   atol=10e-3)


class TestFitSI:

    def setup_method(self, method):
        s = Signal1D(np.random.random((2, 2, 8)))
        m = s.create_model()
        G = Gaussian()
        m.append(G)

        self.model = m
        self.G = G

    def test_fit_spectrum_image(self):
        m = self.model
        G = self.G
        # HyperSpy 2.0: remove setting iterpath='serpentine'
        m.fit_component(G, signal_range=(2, 7), only_current=False,
                        iterpath='serpentine')
        m.axes_manager.indices = (0, 0)
        A = G.A.value
        m.axes_manager.indices = (1, 1)
        B = G.A.value
        assert not A == B



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from operator import attrgetter
from hyperspy.misc.utils import attrsetter
from copy import deepcopy
import dill
from dask.array import Array


def check_that_flags_make_sense(flags):
    # one of: fn, id, sig
    def do_error(f1, f2):
        raise ValueError(
            'The flags "%s" and "%s" are not compatible' %
            (f1, f2))
    if 'fn' in flags:
        if 'id' in flags:
            do_error('fn', 'id')
        if 'sig' in flags:
            do_error('fn', 'sig')
    if 'id' in flags:
        # fn done previously
        if 'sig' in flags:
            do_error('id', 'sig')
        if 'init' in flags:
            do_error('id', 'init')
    # all sig cases already covered


def parse_flag_string(flags):
    return flags.replace(' ', '').split(',')


def export_to_dictionary(target, whitelist, dic, fullcopy=True):
    """ Exports attributes of target from whitelist.keys() to dictionary dic
    All values are references only by default.

    Parameters
    ----------
    target : object
        must contain the (nested) attributes of the whitelist.keys()
    whitelist : dictionary
        A dictionary, keys of which are used as attributes for exporting.
        Key 'self' is only available with tag 'id', when the id of the
        target is saved. The values are either None, or a tuple, where:

        * the first item a string, which containts flags, separated by
          commas.
        * the second item is None if no 'init' flag is given, otherwise
          the object required for the initialization.

        The flag conventions are as follows:

        * 'init': object used for initialization of the target. The object is
          saved in the tuple in whitelist
        * 'fn': the targeted attribute is a function, and may be pickled. A
          tuple of (thing, value) will be exported to the dictionary,
          where thing is None if function is passed as-is, and True if
          dill package is used to pickle the function, with the value as
          the result of the pickle.
        * 'id': the id of the targeted attribute is exported (e.g. id(target.name))
        * 'sig': The targeted attribute is a signal, and will be converted to a
          dictionary if fullcopy=True

    dic : dictionary
        A dictionary where the object will be exported
    fullcopy : bool
        Copies of objects are stored, not references. If any found,
        functions will be pickled and signals converted to dictionaries

    """
    whitelist_flags = {}
    for key, value in whitelist.items():
        if value is None:
            # No flags and/or values are given, just save the target
            thing = attrgetter(key)(target)
            if fullcopy:
                thing = deepcopy(thing)
            dic[key] = thing
            whitelist_flags[key] = ''
            continue

        flags_str, value = value
        flags = parse_flag_string(flags_str)
        check_that_flags_make_sense(flags)
        if key == 'self':
            if 'id' not in flags:
                raise ValueError(
                    'Key "self" is only available with flag "id" given')
            value = id(target)
        else:
            if 'id' in flags:
                value = id(attrgetter(key)(target))

        # here value is either id(thing), or None (all others except 'init'),
        # or something for init
        if 'init' not in flags and value is None:
            value = attrgetter(key)(target)
        # here value either id(thing), or an actual target to export
        if 'sig' in flags:
            if fullcopy:
                from hyperspy.signal import BaseSignal
                if isinstance(value, BaseSignal):
                    value = value._to_dictionary()
                    value['data'] = deepcopy(value['data'])
        elif 'fn' in flags:
            if fullcopy:
                value = (True, dill.dumps(value))
            else:
                value = (None, value)
        elif fullcopy:
            value = deepcopy(value)

        dic[key] = value
        whitelist_flags[key] = flags_str

    if '_whitelist' not in dic:
        dic['_whitelist'] = {}
    # the saved whitelist does not have any values, as they are saved in the
    # original dictionary. Have to restore then when loading from dictionary,
    # most notably all with 'init' flags!!
    dic['_whitelist'].update(whitelist_flags)


def load_from_dictionary(target, dic):
    """ Loads attributes of target to dictionary dic
    The attribute list is read from dic['_whitelist'].keys()

    Parameters
    ----------
        target : object
            must contain the (nested) attributes of the whitelist.keys()
        dic : dictionary
            A dictionary, containing field '_whitelist', which is a dictionary
            with all keys that were exported, with values being flag strings.
            The convention of the flags is as follows:

            * 'init': object used for initialization of the target. Will be 
              copied to the _whitelist after loading
            * 'fn': the targeted attribute is a function, and may have been 
              pickled (preferably with dill package).
            * 'id': the id of the original object was exported and the 
              attribute will not be set. The key has to be '_id_'
            * 'sig': The targeted attribute was a signal, and may have been 
              converted to a dictionary if fullcopy=True

    """
    new_whitelist = {}
    for key, flags_str in dic['_whitelist'].items():
        value = dic[key]
        flags = parse_flag_string(flags_str)
        if 'id' not in flags:
            value = reconstruct_object(flags, value)
            if 'init' in flags:
                new_whitelist[key] = (flags_str, value)
            else:
                attrsetter(target, key, value)
                if len(flags_str):
                    new_whitelist[key] = (flags_str, None)
                else:
                    new_whitelist[key] = None
    if hasattr(target, '_whitelist'):
        if isinstance(target._whitelist, dict):
            target._whitelist.update(new_whitelist)
    else:
        attrsetter(target, '_whitelist', new_whitelist)


def reconstruct_object(flags, value):
    """ Reconstructs the value (if necessary) after having saved it in a
    dictionary
    """
    if not isinstance(flags, list):
        flags = parse_flag_string(flags)
    if 'sig' in flags:
        if isinstance(value, dict):
            from hyperspy.signal import BaseSignal
            value = BaseSignal(**value)
            value._assign_subclass()
        return value
    if 'fn' in flags:
        ifdill, thing = value
        if ifdill is None:
            return thing
        if ifdill in [True, 'True', b'True']:
            return dill.loads(thing)
        # should not be reached
        raise ValueError("The object format is not recognized")
    if isinstance(value, Array):
        value = value.compute()
    return value



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


def set_log_level(level):
    """
    Convenience function to set the log level of all hyperspy modules.

    Note: The log level of all other modules are left untouched.

    Parameters
    ----------
    level : int or str
        The log level to set. Any values that `logging.Logger.setLevel()`
        accepts are valid. The default options are:

        - 'CRITICAL'
        - 'ERROR'
        - 'WARNING'
        - 'INFO'
        - 'DEBUG'
        - 'NOTSET'

    Example
    -------
    For normal logging of hyperspy functions, you can set the log level like
    this:

    >>> import hyperspy.api as hs
    >>> hs.set_log_level('INFO')
    >>> hs.load(r'my_file.dm3')
    INFO:hyperspy.io_plugins.digital_micrograph:DM version: 3
    INFO:hyperspy.io_plugins.digital_micrograph:size 4796607 B
    INFO:hyperspy.io_plugins.digital_micrograph:Is file Little endian? True
    INFO:hyperspy.io_plugins.digital_micrograph:Total tags in root group: 15
    <Signal2D, title: My file, dimensions: (|1024, 1024)>

    If you need the log output during the initial import of hyperspy, you
    should set the log level like this:

    >>> from hyperspy.logger import set_log_level
    >>> set_log_level('DEBUG')
    >>> import hyperspy.api as hs
    DEBUG:hyperspy.gui:Loading hyperspy.gui
    DEBUG:hyperspy.gui:Current MPL backend: TkAgg
    DEBUG:hyperspy.gui:Current ETS toolkit: qt4
    DEBUG:hyperspy.gui:Current ETS toolkit set to: null

    """
    import logging
    logging.basicConfig()  # Does nothing if already configured
    logging.getLogger('hyperspy').setLevel(level)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging
from multiprocessing import cpu_count

import dill
import numpy as np

from hyperspy.misc.utils import DictionaryTreeBrowser
from hyperspy.misc.utils import slugify
from hyperspy.misc.math_tools import check_random_state
from hyperspy.external.progressbar import progressbar
from hyperspy.signal import BaseSignal
from hyperspy.samfire_utils.strategy import (LocalStrategy,
                                             GlobalStrategy)
from hyperspy.samfire_utils.local_strategies import ReducedChiSquaredStrategy
from hyperspy.samfire_utils.global_strategies import HistogramStrategy


_logger = logging.getLogger(__name__)


class StrategyList(list):

    def __init__(self, samf):
        super(StrategyList, self).__init__()
        self.samf = samf

    def append(self, thing):
        thing.samf = self.samf
        list.append(self, thing)

    def extend(self, iterable):
        for thing in iterable:
            self.append(thing)

    def remove(self, thing):
        if isinstance(thing, int):
            thing = self[thing]
        thing.samf = None
        list.remove(self, thing)

    def __repr__(self):
        signature = u"%3s | %4s | %s"
        ans = signature % ("A", "#", "Strategy")
        ans += u"\n"
        ans += signature % (u'-' * 2, u'-' * 4, u'-' * 25)
        if self:
            for n, s in enumerate(self):
                ans += u"\n"
                name = repr(s)
                a = u" x" if self.samf._active_strategy_ind == n else u""
                ans += signature % (a, str(n), name)
        return ans


class Samfire:

    """Smart Adaptive Multidimensional Fitting (SAMFire) object

    SAMFire is a more robust way of fitting multidimensional datasets. By
    extracting starting values for each pixel from already fitted pixels,
    SAMFire stops the fitting algorithm from getting lost in the parameter
    space by always starting close to the optimal solution.

    SAMFire only picks starting parameters and the order the pixels (in the
    navigation space) are fitted, and does not provide any new minimisation
    algorithms.

    Attributes
    ----------

    model : Model instance
        The complete model
    optional_components : list
        A list of components that can be switched off at some pixels if it
        returns a better Akaike's Information Criterion with correction (AICc)
    workers : int
        A number of processes that will perform the fitting parallely
    pool : samfire_pool instance
        A proxy object that manages either multiprocessing or ipyparallel pool
    strategies : strategy list
        A list of strategies that will be used to select pixel fitting order
        and calculate required starting parameters. Strategies come in two
        "flavours" - local and global. Local strategies spread the starting
        values to the nearest pixels and forces certain pixel fitting order.
        Global strategies look for clusters in parameter values, and suggests
        most frequent values. Global strategy do not depend on pixel fitting
        order, hence it is randomised.
    metadata : dictionary
        A dictionary for important samfire parameters
    active_strategy : strategy
        The currently active strategy from the strategies list
    update_every : int
        If segmenter strategy is running, updates the historams every time
        update_every good fits are found.
    plot_every : int
        When running, samfire plots results every time plot_every good fits are
        found.
    save_every : int
        When running, samfire saves results every time save_every good fits are
        found.
    random_state : None or int or RandomState instance, default None
        Random seed used to select the next pixels.

    Methods
    -------

    start
        start SAMFire
    stop
        stop SAMFire
    plot
        force plot of currently selected active strategy
    refresh_database
        refresh current active strategy database. No previous structure is
        preserved
    backup
        backs up the current version of the model
    change_strategy
        changes strategy to a new one. Certain rules apply
    append
        appends strategy to the strategies list
    extend
        extends strategies list
    remove
        removes strategy from strategies list
    update
        updates the current model with values, received from a worker
    log
        if _log exists, logs the arguments to the list.
    generate_values
        creates a generator to calculate values to be sent to the workers
    """

    __active_strategy_ind = 0
    _progressbar = None
    pool = None
    _figure = None
    optional_components = []
    running_pixels = []
    plot_every = 0
    save_every = np.nan
    _workers = None
    _args = None
    count = 0

    def __init__(self, model, workers=None, setup=True, random_state=None, **kwargs):
        # constants:
        if workers is None:
            workers = max(1, cpu_count() - 1)
        self.model = model
        self.metadata = DictionaryTreeBrowser()

        self._scale = 1.0
        # -1 -> done pixel, use
        # -2 -> done, ignore when diffusion
        #  0 -> bad fit/no info
        # >0 -> select when turn comes

        self.metadata.add_node('marker')
        self.metadata.add_node('goodness_test')

        marker = np.empty(self.model.axes_manager.navigation_shape[::-1])
        marker.fill(self._scale)

        self.metadata.marker = marker
        self.strategies = StrategyList(self)
        self.strategies.append(ReducedChiSquaredStrategy())
        self.strategies.append(HistogramStrategy())
        self._active_strategy_ind = 0
        self.update_every = max(10, workers * 2)  # some sensible number....
        from hyperspy.samfire_utils.fit_tests import red_chisq_test
        self.metadata.goodness_test = red_chisq_test(tolerance=1.0)
        self.metadata._gt_dump = None
        from hyperspy.samfire_utils.samfire_kernel import single_kernel
        self.single_kernel = single_kernel
        self._workers = workers
        if len(kwargs) or setup:
            self._setup(**kwargs)
        self.refresh_database()
        self.random_state = check_random_state(random_state)

    @property
    def active_strategy(self):
        """Returns the active strategy"""
        return self.strategies[self._active_strategy_ind]

    @active_strategy.setter
    def active_strategy(self, value):
        self.change_strategy(value)

    def _setup(self, **kwargs):
        """Set up SAMFire - configure models, set up pool if necessary"""
        from hyperspy.samfire_utils.samfire_pool import SamfirePool
        self._figure = None
        self.metadata._gt_dump = dill.dumps(self.metadata.goodness_test)
        self._enable_optional_components()

        if hasattr(self.model, '_suspend_auto_fine_structure_width'):
            self.model._suspend_auto_fine_structure_width = True

        if hasattr(self, '_log'):
            self._log = []

        if self._workers and self.pool is None:
            if 'num_workers' not in kwargs:
                kwargs['num_workers'] = self._workers
            if self.pool is None:
                self.pool = SamfirePool(**kwargs)
            self._workers = self.pool.num_workers
            self.pool.prepare_workers(self)

    def start(self, **kwargs):
        """Starts SAMFire.

        Parameters
        ----------
        **kwargs : key-word arguments
            Any key-word arguments to be passed to Model.fit() call
        """
        self._setup()
        if self._workers and self.pool is not None:
            self.pool.update_parameters()
        if 'min_function' in kwargs:
            kwargs['min_function'] = dill.dumps(kwargs['min_function'])
        if 'min_function_grad' in kwargs:
            kwargs['min_function_grad'] = dill.dumps(
                kwargs['min_function_grad'])
        self._args = kwargs
        num_of_strat = len(self.strategies)
        total_size = self.model.axes_manager.navigation_size - self.pixels_done
        self._progressbar = progressbar(total=total_size)
        try:
            while True:
                self._run_active_strategy()
                self.plot()
                if self.pixels_done == self.model.axes_manager.navigation_size:
                    # all pixels are done, no need to go to the next strategy
                    break
                if self._active_strategy_ind == num_of_strat - 1:
                    # last one just finished running
                    break
                self.change_strategy(self._active_strategy_ind + 1)
        except KeyboardInterrupt:  # pragma: no cover
            if self.pool is not None:
                _logger.warning(
                    'Collecting already started pixels, please wait')
                self.pool.collect_results()

    def append(self, strategy):
        """appends the given strategy to the end of the strategies list

        Parameters
        ----------
        strategy : strategy instance
        """
        self.strategies.append(strategy)

    def extend(self, iterable):
        """extend the strategies list by the given iterable

        Parameters
        ----------
        iterable : an iterable of strategy instances
        """
        self.strategies.extend(iterable)

    def remove(self, thing):
        """removes given strategy from the strategies list

        Parameters
        ----------
        thing : int or strategy instance
            Strategy that is in current strategies list or its index.
        """
        self.strategies.remove(thing)

    @property
    def _active_strategy_ind(self):
        return self.__active_strategy_ind

    @_active_strategy_ind.setter
    def _active_strategy_ind(self, value):
        self.__active_strategy_ind = np.abs(int(value))

    def _run_active_strategy(self):
        if self.pool is not None:
            self.count = 0
            self.pool.run()
        else:
            self._run_active_strategy_one()

    @property
    def pixels_left(self):
        """Returns the number of pixels that are left to solve. This number can
        increase as SAMFire learns more information about the data.
        """
        return np.sum(self.metadata.marker > 0.)

    @property
    def pixels_done(self):
        """Returns the number of pixels that have been solved"""
        return np.sum(self.metadata.marker <= -self._scale)

    def _run_active_strategy_one(self):
        self.count = 0
        while self.pixels_left:
            ind = self._next_pixels(1)[0]
            vals = self.active_strategy.values(ind)
            self.running_pixels.append(ind)
            isgood = self.single_kernel(self.model,
                                        ind,
                                        vals,
                                        self.optional_components,
                                        self._args,
                                        self.metadata.goodness_test)
            self.running_pixels.remove(ind)
            self.count += 1
            if isgood:
                self._progressbar.update(1)
            self.active_strategy.update(ind, isgood)
            self.plot(on_count=True)
            self.backup(on_count=True)

    def backup(self, filename=None, on_count=True):
        """Backs-up the samfire results in a file

        Parameters
        ----------
        filename: {str, None}
            the filename. If None, a default value of "backup_"+signal_title
            is used.
        on_count: bool
            if True (default), only saves on the required count of steps
        """
        if filename is None:
            title = self.model.signal.metadata.General.title
            filename = slugify('backup_' + title)
        # maybe add saving marker + strategies as well?
        if self.count % self.save_every == 0 or not on_count:
            self.model.save(filename,
                            name='samfire_backup', overwrite=True)
            self.model.signal.models.remove('samfire_backup')

    def update(self, ind, results=None, isgood=None):
        """Updates the current model with the results, received from the
        workers. Results are only stored if the results are good enough

        Parameters
        ----------
        ind : tuple
            contains the index of the pixel of the results
        results : {dict, None}
            dictionary of the results. If None, means we are updating in-place
            (e.g. refreshing the marker or strategies)
        isgood : {bool, None}
            if it is known if the results are good according to the
            goodness-of-fit test. If None, the pixel is tested
        """
        if results is not None and (isgood is None or isgood):
            self._swap_dict_and_model(ind, results)

        if isgood is None:
            isgood = self.metadata.goodness_test.test(self.model, ind)
        self.count += 1
        if isgood and self._progressbar is not None:
            self._progressbar.update(1)

        self.active_strategy.update(ind, isgood)
        if not isgood and results is not None:
            self._swap_dict_and_model(ind, results)

    def refresh_database(self):
        """Refreshes currently selected strategy without preserving any
        "ignored" pixels
        """
        # updates current active strategy database / prob.
        # Assume when chisq is not None, it's relevant

        # TODO: if no calculated pixels, request user input

        calculated_pixels = np.logical_not(np.isnan(self.model.red_chisq.data))
        # only include pixels that are good enough
        calculated_pixels = self.metadata.goodness_test.map(
            self.model,
            calculated_pixels)

        self.active_strategy.refresh(True, calculated_pixels)

    def change_strategy(self, new_strat):
        """Changes current strategy to a new one. Certain rules apply:
        diffusion -> diffusion : resets all "ignored" pixels
        diffusion -> segmenter : saves already calculated pixels to be ignored
        when(if) subsequently diffusion strategy is run

        Parameters
        ----------
        new_strat : {int | strategy}
            index of the new strategy from the strategies list or the
            strategy object itself
        """
        from numbers import Number
        if not isinstance(new_strat, Number):
            try:
                new_strat = self.strategies.index(new_strat)
            except ValueError:
                raise ValueError(
                    "The passed object is not in current strategies list")

        new_strat = np.abs(int(new_strat))
        if new_strat == self._active_strategy_ind:
            self.refresh_database()

        # copy previous "done" pixels to the new one, delete old database

        # TODO: make sure it's a number. Get index if object is passed?
        if new_strat >= len(self.strategies):
            raise ValueError('too big new strategy index')

        current = self.active_strategy
        new = self.strategies[new_strat]

        if isinstance(current, LocalStrategy) and isinstance(
                new, LocalStrategy):
            # forget ignore/done levels, keep just calculated or not
            new.refresh(True)
        else:
            if isinstance(current, LocalStrategy) and isinstance(
                    new, GlobalStrategy):
                # if diffusion->segmenter, set previous -1 to -2 (ignored for
                # the next diffusion)
                self.metadata.marker[
                    self.metadata.marker == -
                    self._scale] -= self._scale

            new.refresh(False)
        current.clean()
        if current.close_plot is not None:
            current.close_plot()
        self._active_strategy_ind = new_strat

    def generate_values(self, need_inds):
        """Returns an iterator that yields the index of the pixel and the
        value dictionary to be sent to the workers.

        Parameters
        ----------
        need_inds: int
            the number of pixels to be returned in the generator
        """
        if need_inds:
            # get pixel index
            for ind in self._next_pixels(need_inds):
                # get starting parameters / array of possible values
                value_dict = self.active_strategy.values(ind)
                value_dict['fitting_kwargs'] = self._args
                value_dict['signal.data'] = \
                    self.model.signal.data[ind + (...,)]
                if self.model.signal._lazy:
                    value_dict['signal.data'] = value_dict[
                        'signal.data'].compute()
                if self.model.signal.metadata.has_item(
                        'Signal.Noise_properties.variance'):
                    var = self.model.signal.metadata.Signal.Noise_properties.variance
                    if isinstance(var, BaseSignal):
                        dat = var.data[ind + (...,)]
                        value_dict['variance.data'] = dat.compute(
                        ) if var._lazy else dat
                if hasattr(self.model,
                           'low_loss') and self.model.low_loss is not None:
                    dat = self.model.low_loss.data[ind + (...,)]
                    value_dict['low_loss.data'] = dat.compute(
                    ) if self.model.low_loss._lazy else dat

                self.running_pixels.append(ind)
                self.metadata.marker[ind] = 0.
                yield ind, value_dict

    def _next_pixels(self, number):
        best = self.metadata.marker.max()
        inds = []
        if best > 0.0:
            ind_list = np.where(self.metadata.marker == best)
            while number and ind_list[0].size > 0:
                i = self.random_state.randint(len(ind_list[0]))
                ind = tuple([lst[i] for lst in ind_list])
                if ind not in self.running_pixels:
                    inds.append(ind)
                # removing the added indices
                ind_list = [np.delete(lst, i, 0) for lst in ind_list]
                number -= 1
        return inds

    def _swap_dict_and_model(self, m_ind, dict_, d_ind=None):
        if d_ind is None:
            d_ind = tuple([0 for _ in dict_['dof.data'].shape])
        m = self.model
        for k in dict_.keys():
            if k.endswith('.data'):
                item = k[:-5]
                getattr(m, item).data[m_ind], dict_[k] = \
                    dict_[k].copy(), getattr(m, item).data[m_ind].copy()
        for comp_name, comp in dict_['components'].items():
            # only active components are sent
            if self.model[comp_name].active_is_multidimensional:
                self.model[comp_name]._active_array[m_ind] = True
            self.model[comp_name].active = True

            for param_model in self.model[comp_name].parameters:
                param_dict = comp[param_model.name]
                param_model.map[m_ind], param_dict[d_ind] = \
                    param_dict[d_ind].copy(), param_model.map[m_ind].copy()

        for component in self.model:
            # switch off all that did not appear in the dictionary
            if component.name not in dict_['components'].keys():
                if component.active_is_multidimensional:
                    component._active_array[m_ind] = False

    def _enable_optional_components(self):
        if len(self.optional_components) == 0:
            return
        for c in self.optional_components:
            comp = self.model._get_component(c)
            if not comp.active_is_multidimensional:
                comp.active_is_multidimensional = True
        if not np.all([isinstance(a, int) for a in
                       self.optional_components]):
            new_list = []
            for op in self.optional_components:
                for ic, c in enumerate(self.model):
                    if c is self.model._get_component(op):
                        new_list.append(ic)
            self.optional_components = new_list

    def _request_user_input(self):
        from hyperspy.signals import Image
        from hyperspy.drawing.widgets import SquareWidget
        mark = Image(self.metadata.marker,
                     axes=self.model.axes_manager._get_navigation_axes_dicts())
        mark.metadata.General.title = 'SAMFire marker'

        def update_when_triggered():
            ind = self.model.axes_manager.indices[::-1]
            isgood = self.metadata.goodness_test.test(self.model, ind)
            self.active_strategy.update(ind, isgood, 0)
            mark.events.data_changed.trigger(mark)

        self.model.plot()
        self.model.events.fitted.connect(update_when_triggered, [])
        self.model._plot.signal_plot.events.closed.connect(
            lambda: self.model.events.fitted.disconnect(update_when_triggered),
            [])

        mark.plot(navigator='slider')

        w = SquareWidget(self.model.axes_manager)
        w.color = 'yellow'
        w.set_mpl_ax(mark._plot.signal_plot.ax)
        w.connect_navigate()

        def connect_other_navigation1(axes_manager):
            with mark.axes_manager.events.indices_changed.suppress_callback(
                    connect_other_navigation2):
                for ax1, ax2 in zip(mark.axes_manager.navigation_axes,
                                    axes_manager.navigation_axes[2:]):
                    ax1.value = ax2.value

        def connect_other_navigation2(axes_manager):
            with self.model.axes_manager.events.indices_changed.suppress_callback(
                    connect_other_navigation1):
                for ax1, ax2 in zip(self.model.axes_manager.navigation_axes[2:],
                                    axes_manager.navigation_axes):
                    ax1.value = ax2.value

        mark.axes_manager.events.indices_changed.connect(
            connect_other_navigation2, {'obj': 'axes_manager'})
        self.model.axes_manager.events.indices_changed.connect(
            connect_other_navigation1, {'obj': 'axes_manager'})

        self.model._plot.signal_plot.events.closed.connect(
            lambda: mark._plot.close, [])
        self.model._plot.signal_plot.events.closed.connect(
            lambda: self.model.axes_manager.events.indices_changed.disconnect(
                connect_other_navigation1), [])

    def plot(self, on_count=False):
        """(if possible) plots current strategy plot. Local strategies plot
        grayscale navigation signal with brightness representing order of the
        pixel selection. Global strategies plot a collection of histograms,
        one per parameter.

        Parameters
        ----------
        on_count : bool
            if True, only tries to plot every speficied count, otherwise
            (default) always plots if possible.
        """
        count_test = self.plot_every and (self.count % self.plot_every == 0)
        if not on_count or count_test:
            if self.strategies:
                try:
                    self._figure = self.active_strategy.plot(self._figure)
                except BaseException:
                    self._figure = None
                    self._figure = self.active_strategy.plot(self._figure)

    def log(self, *args):
        """If has a list named "_log", appends the arguments there
        """
        if hasattr(self, '_log') and isinstance(self._log, list):
            self._log.append(args)

    def __repr__(self):
        ans = u"<SAMFire of the signal titled: '"
        ans += self.model.signal.metadata.General.title
        ans += u"'>"
        return ans

    def stop(self):
        if hasattr(self, "pool") and self.pool is not None:
            self.pool.stop()



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


from hyperspy.exceptions import DataDimensionError
from hyperspy.docstrings.signal import OPTIMIZE_ARG


class CommonSignal1D(object):

    """Common functions for 1-dimensional signals."""

    def to_signal2D(self, optimize=True):
        """Returns the one dimensional signal as a two dimensional signal.

        By default ensures the data is stored optimally, hence often making a
        copy of the data. See `transpose` for a more general method with more
        options.

        %s

        See Also
        --------
        transpose, as_signal1D, as_signal2D, hs.transpose

        Raises
        ------
        DataDimensionError
            When data.ndim < 2


        """
        if self.data.ndim < 2:
            raise DataDimensionError(
                "A Signal dimension must be >= 2 to be converted to Signal2D")
        nat = self.axes_manager._get_axes_in_natural_order()
        im = self.transpose(signal_axes=nat[:2], navigation_axes=nat[2:],
                            optimize=optimize)
        return im
    to_signal2D.__doc__ %= (OPTIMIZE_ARG.replace('False', 'True'))



# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import Mock

import numpy as np
import matplotlib
import pytest

from hyperspy.drawing.utils import create_figure, contrast_stretching


def test_create_figure():
    if matplotlib.get_backend() == "agg":
        pytest.xfail("{} backend does not support on_close event.".format(
            matplotlib.get_backend()))

    dummy_function = Mock()
    fig = create_figure(window_title="test title",
                        _on_figure_window_close=dummy_function)
    assert isinstance(fig, matplotlib.figure.Figure) == True
    matplotlib.pyplot.close(fig)
    dummy_function.assert_called_once_with()


def test_contrast_stretching():
    data = np.arange(100)
    assert contrast_stretching(data, 1, 99) == (1, 99)
    assert contrast_stretching(data, 1.0, 99.0) == (1, 99)
    assert contrast_stretching(data, '1th', '99th') == (0.99, 98.01)
    assert contrast_stretching(data, '0.05th', '99.95th') == (0.0495, 98.9505)
    # vmin, vmax are to set in conftest.py
    assert contrast_stretching(data, None, None) == (0.0, 99.0)



"""Nexus file reading, writing and inspection."""
# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
import numpy as np
import dask.array as da
import os
import h5py
import pprint
import traits.api as t
from hyperspy.io_plugins.hspy import overwrite_dataset, get_signal_chunks
from hyperspy.misc.utils import DictionaryTreeBrowser
_logger = logging.getLogger(__name__)
# Plugin characteristics

format_name = 'Nexus'
description = \
    'Read NXdata sets from Nexus files and metadata. Data and metadata can '\
    'also be examined from general hdf5 files'
full_support = False
# Recognised file extension
file_extensions = ['nxs', 'NXS']
default_extension = 0
# Writing capabilities:
writes = True


def _byte_to_string(value):
    """Decode a byte string.

    Parameters
    ----------
    value :  byte str

    Returns
    -------
    str
        decoded version of input value

    """
    try:
        text = value.decode("utf-8")
    except UnicodeDecodeError:
        text = value.decode('latin-1')
    return text.replace('\x00', '').rstrip()


def _parse_from_file(value, lazy=False):
    """To convert values from the hdf file to compatible formats.

    When reading string arrays we convert or keep string arrays as
    byte_strings (some io_plugins only supports byte-strings arrays so this
    ensures inter-compatibility across io_plugins)
    Arrays of length 1 - return the single value stored.
    Large datasets are returned as dask arrays if lazy=True.

    Parameters
    ----------
    value : input read from hdf file (array,list,tuple,string,int,float)
    lazy  : bool  {default: False}
        The lazy flag is only applied to values of size >=2

    Returns
    -------
    str,int, float, ndarray dask Array
        parsed value.

    """
    toreturn = value
    if isinstance(value, h5py.Dataset):
        if value.size < 2:
            toreturn = value[...].item()
        else:
            if lazy:
                if value.chunks:
                    toreturn = da.from_array(value, value.chunks)
                else:
                    chunks = get_signal_chunks(value.shape, value.dtype)
                    toreturn = da.from_array(value, chunks)
            else:
                toreturn = np.array(value)

    if isinstance(toreturn, np.ndarray) and value.shape == (1,):
        toreturn = toreturn[0]
    if isinstance(toreturn, bytes):
        toreturn = _byte_to_string(toreturn)
    if isinstance(toreturn, (np.int, np.float)):
        toreturn = toreturn
    if isinstance(toreturn, (np.ndarray)) and toreturn.dtype.char == "U":
        toreturn = toreturn.astype("S")
    return toreturn


def _parse_to_file(value):
    """Convert to a suitable format for writing to HDF5.

    For example unicode values are not compatible with hdf5 so conversion to
    byte strings is required.

    Parameters
    ----------
    value - input object to write to the hdf file

    Returns
    -------
    parsed value

    """
    totest = value
    toreturn = totest
    if isinstance(totest, (bytes, int, float)):
        toreturn = value
    if isinstance(totest, (list, tuple)):
        totest = np.array(value)
    if isinstance(totest, np.ndarray) and totest.dtype.char == "U":
        toreturn = np.array(totest).astype("S")
    elif isinstance(totest, (np.ndarray, da.Array)):
        toreturn = totest
    if isinstance(totest, str):
        toreturn = totest.encode("utf-8")
        toreturn = np.string_(toreturn)
    return toreturn


def _text_split(s, sep):
    """Split a string based of list of seperators.

    Parameters
    ----------
    s   : str
    sep : str  - seperator or list of seperators e.g. '.' or ['_','/']

    Returns
    -------
    list
       String sections split based on the seperators

    """
    stack = [s]
    for char in sep:
        pieces = []
        for substr in stack:
            pieces.extend(substr.split(char))
        stack = pieces
    if '' in stack:
        stack.remove('')
    return stack


def _getlink(h5group, rootkey, key):
    """Return the link target path.

    If a hdf group is a soft link or has a target attribute
    this method will return the target path. If no link is found
    return None.

    Returns
    -------
    str
        Soft link path if it exists, otherwise None

    """
    _target = None
    if rootkey != '/':
        if isinstance(h5group, h5py.Group):
            _link = h5group.get(key, getlink=True)
            if isinstance(_link, h5py.SoftLink):
                _target = _link.path
        if 'target' in h5group.attrs.keys():
            _target = _parse_from_file(h5group.attrs['target'])
            if not _target.startswith('/'):
                _target = '/' + _target
            if _target == rootkey:
                _target = None

    return _target


def _extract_hdf_dataset(group, dataset, lazy=False):
    """Import data from hdf path.

    Parameters
    ----------
    group : hdf group
        group from which to load the dataset
    dataset : str
        path to the dataset within the group
    lazy    : bool {default:True}
        If true use lazy opening, if false read into memory

    Returns
    -------
    dask or numpy array

    """
    data = group[dataset]
    if lazy:
        if "chunks" in data.attrs.keys():
            chunks = data.attrs["chunks"]
        else:
            chunks = get_signal_chunks(data.shape, data.dtype)
        data_lazy = da.from_array(data, chunks=chunks)
    else:
        data_lazy = np.array(data)

    nav_list = []
    for i in range(data.ndim):
        nav_list.append({
                'size': data.shape[i],
                'index_in_array': i,
                'scale': 1,
                'offset': 0.0,
                'units': '',
                'navigate': True,
               })

    dictionary = {'data': data_lazy, 'metadata': {}, 'original_metadata': {},
                  'axes': nav_list}

    return dictionary


def _nexus_dataset_to_signal(group, nexus_dataset_path, lazy=False):
    """Load an NXdata set as a hyperspy signal.

    Parameters
    ----------
    group : hdf group containing the NXdata
    nexus_data_path : str
        Path to the NXdata set in the group
    lazy : bool, default : True
        lazy loading of data

    Returns
    -------
    dict
        A signal dictionary which can be used to instantiate a signal.

    """
    detector_index = 0
    interpretation = None
    dataentry = group[nexus_dataset_path]
    if "signal" in dataentry.attrs.keys():
        if _is_int(dataentry.attrs["signal"]):
            data_key = "data"
        else:
            data_key = dataentry.attrs["signal"]
    else:
        _logger.info("No signal attr associated with NXdata will\
                     try assume signal name is data")
        if "data" not in dataentry.keys():
            raise ValueError("Signal attribute not found in NXdata and\
                             attempt to find a default \"data\" key failed")
        else:
            data_key = "data"

    if "interpretation" in dataentry.attrs.keys():
        interpretation = _parse_from_file(dataentry.attrs["interpretation"])

    data = dataentry[data_key]
    nav_list = []
    # list indices...
    axis_index_list = []
    if "axes" in dataentry.attrs.keys():
        axes_key = dataentry.attrs["axes"]
        axes_list = ["."]*data.ndim
        if isinstance(axes_key, np.ndarray):
            for i, num in enumerate(axes_key):
                axes_list[i] = _parse_from_file(num)
        else:
            axes_list[0] = _parse_from_file(axes_key)

        named_axes = list(range(len(axes_list)))
        for i, ax in enumerate(axes_list):
            if ax != ".":
                index_name = ax + "_indices"
                if index_name in dataentry.attrs:
                    ind_in_array = int(dataentry.attrs[index_name])
                else:
                    ind_in_array = i
                axis_index_list.append(ind_in_array)
                if "units" in dataentry[ax].attrs:
                    units = _parse_from_file(dataentry[ax].attrs["units"])
                else:
                    units = ""

                navigation = True
                named_axes.remove(ind_in_array)

                if _is_numeric_data(dataentry[ax]):
                    if dataentry[ax].size > 1:
                        if _is_linear_axis(dataentry[ax]):
                            nav_list.append({
                                'size': data.shape[ind_in_array],
                                'index_in_array': ind_in_array,
                                'name': ax,
                                'scale': abs(dataentry[ax][1] -
                                             dataentry[ax][0]),
                                'offset': min(dataentry[ax][0],
                                              dataentry[ax][-1]),
                                'units': units,
                                'navigate': navigation
                                })
                        else:
                            nav_list.append({
                                'size': data.shape[ind_in_array],
                                'index_in_array': ind_in_array,
                                'name': ax,
                                'scale': 1,
                                'offset': 0,
                                'navigate': navigation
                                })
                    else:
                        nav_list.append({
                            'size': 1,
                            'index_in_array': ind_in_array,
                            'name': ax,
                            'scale': 1,
                            'offset': dataentry[ax][0],
                            'units': units,
                            'navigate': True
                            })
            else:
                if len(data.shape) == len(axes_list):
                    nav_list.append({
                            'size': data.shape[named_axes[detector_index]],
                            'index_in_array': named_axes[detector_index],
                            'scale': 1,
                            'offset': 0.0,
                            'units': '',
                            'navigate': False
                           })
                    detector_index = detector_index+1

    if lazy:
        if "chunks" in data.attrs.keys():
            chunks = data.attrs["chunks"]
        else:
            chunks = get_signal_chunks(data.shape, data.dtype)
        data_lazy = da.from_array(data, chunks=chunks)
    else:
        data_lazy = np.array(data)

    if not nav_list:
        for i in range(data.ndim):
            nav_list.append({
                    'size': data_lazy.shape[i],
                    'index_in_array': i,
                    'scale': 1,
                    'offset': 0.0,
                    'units': '',
                    'navigate': True
                   })
    title = _text_split(nexus_dataset_path, '/')[-1]
    metadata = {'General': {'title': title}}

    #
    # if interpretation - reset the nav axes
    # assume the last dimensions are the signal
    #
    if interpretation:
        for x in nav_list:
            x["navigate"] = True
        if interpretation == "spectrum":
            nav_list[-1]["navigate"] = False
        elif interpretation == "image":
            nav_list[-1]["navigate"] = False
            nav_list[-2]["navigate"] = False

    dictionary = {'data': data_lazy,
                  'axes': nav_list,
                  'metadata': metadata}
    return dictionary


def file_reader(filename, lazy=False, dataset_keys=None,
                metadata_keys=None,
                nxdata_only=False,
                hardlinks_only=False,
                use_default=False,
                **kwds):
    """Read NXdata class or hdf datasets from a file and return signal(s).

    Note
    ----
    Loading all datasets can result in a large number of signals
    Please review your datasets and use the dataset_keys to target
    the datasets of interest.
    "keys" is a special keywords and prepended with "fix" in the metadata
    structure to avoid any issues.

    Datasets are all arrays with size>2 (arrays, lists)

    Parameters
    ----------
    filename : str
        Input filename
    dataset_keys  : None, str, list of strings, default : None
        If None all datasets are returned.
        If a string or list of strings is provided only items
        whose path contain the string(s) are returned. For example
        dataset_keys = ["instrument", "Fe"] will return
        data entries with instrument or Fe in their hdf path.
    metadata_keys: : None, str, list of strings, default : None
        Only return items from the original metadata whose path contain the
        strings .e.g metadata_keys = ["instrument", "Fe"] will return
        all metadata entries with "instrument" or "Fe" in their hdf path.
    nxdata_only : bool, default : False
        If True only NXdata will be converted into a signal
        if False NXdata and any hdf datasets will be loaded as signals
    hardlinks_only : bool, default : False
        If True any links (soft or External) will be ignored when loading.
    use_default : bool, default : False
        If True and a default NXdata is defined in the file load this as a
        signal. This will ignore the other keyword options. If True and no
        default is defined the file will be loaded according to
        the keyword options.

    Returns
    -------
    dict : signal dictionary or list of signal dictionaries


    See Also
    --------
    * :py:meth:`~.io_plugins.nexus.list_datasets_in_file`
    * :py:meth:`~.io_plugins.nexus.read_metadata_from_file`


    """
    # search for NXdata sets...

    mapping = kwds.get('mapping', {})
    original_metadata = {}
    learning = {}
    fin = h5py.File(filename, "r")
    signal_dict_list = []

    dataset_keys = _check_search_keys(dataset_keys)
    metadata_keys = _check_search_keys(metadata_keys)
    original_metadata = _load_metadata(fin, lazy=lazy)
    # some default values...
    nexus_data_paths = []
    hdf_data_paths = []
    # check if a default dataset is defined
    if use_default:
        nexus_data_paths, hdf_data_paths = _find_data(fin,
                                                      search_keys=None,
                                                      hardlinks_only=False)
        nxentry = None
        nxdata = None
        if "attrs" in original_metadata:
            if "default" in original_metadata["attrs"]:
                nxentry = original_metadata["attrs"]["default"]
            else:
                rootlist = list(original_metadata.keys())
                rootlist.remove("attrs")
                if rootlist and len(rootlist) == 1:
                    nxentry == rootlist[0]
            if nxentry:
                if "default" in original_metadata[nxentry]["attrs"]:
                    nxdata = original_metadata[nxentry]["attrs"]["default"]
            if nxentry and nxdata:
                nxdata = "/"+nxentry+"/"+nxdata
            if nxdata:
                hdf_data_paths = []
                nexus_data_paths = [nxpath for nxpath in nexus_data_paths
                                    if nxdata in nxpath]
    # if no default found then search for the data as normal
    if not nexus_data_paths and not hdf_data_paths:
        nexus_data_paths, hdf_data_paths = \
            _find_data(fin, search_keys=dataset_keys,
                       hardlinks_only=hardlinks_only)

    for data_path in nexus_data_paths:
        dictionary = _nexus_dataset_to_signal(fin, data_path, lazy=lazy)
        entryname = _text_split(data_path, "/")[0]
        dictionary["mapping"] = mapping
        title = dictionary["metadata"]["General"]["title"]
        if entryname in original_metadata:
            if metadata_keys is None:
                dictionary["original_metadata"] = \
                    original_metadata[entryname]
            else:
                dictionary["original_metadata"] = \
                    _find_search_keys_in_dict(original_metadata,
                                              search_keys=metadata_keys)
            # test if it's a hyperspy_nexus format and update metadata
            # as appropriate.
            if "attrs" in original_metadata and \
                    "file_writer" in original_metadata["attrs"]:
                if original_metadata["attrs"]["file_writer"] == \
                        "hyperspy_nexus_v3":
                    orig_metadata = original_metadata[entryname]
                    if "auxiliary" in orig_metadata:
                        if "learning_results" in orig_metadata["auxiliary"]:
                            learning = \
                                orig_metadata["auxiliary"]["learning_results"]
                            dictionary["attributes"] = {}
                            dictionary["attributes"]["learning_results"] = \
                                learning
                        if "original_metadata" in orig_metadata["auxiliary"]:
                            if metadata_keys is None:
                                dictionary["original_metadata"] = \
                                    (orig_metadata["auxiliary"]
                                     ["original_metadata"])
                            else:
                                dictionary["original_metadata"] = \
                                    _find_search_keys_in_dict(
                                        (orig_metadata["auxiliary"]
                                         ["original_metadata"]),
                                        search_keys=metadata_keys)
                        if "hyperspy_metadata" in orig_metadata["auxiliary"]:
                            hyper_metadata = \
                                orig_metadata["auxiliary"]["hyperspy_metadata"]
                            hyper_metadata.update(dictionary["metadata"])
                            dictionary["metadata"] = hyper_metadata
        else:
            dictionary["original_metadata"] = {}

        signal_dict_list.append(dictionary)

    if not nxdata_only:
        for data_path in hdf_data_paths:
            datadict = _extract_hdf_dataset(fin, data_path, lazy=lazy)
            if datadict:
                title = data_path[1:].replace('/', '_')
                basic_metadata = {'General':
                                  {'original_filename':
                                   os.path.split(filename)[1],
                                   'title': title}}
                datadict["metadata"].update(basic_metadata)
                signal_dict_list.append(datadict)

    return signal_dict_list


def _is_linear_axis(data):
    """Check if the data is linearly incrementing.

    Parameter
    ---------
    data : dask or numpy array

    Returns
    -------
    bool
       True or False

    """
    steps = np.diff(data)
    est_steps = np.array([steps[0]]*len(steps))
    return np.allclose(est_steps, steps, rtol=1.0e-5)


def _is_numeric_data(data):
    """Check that data contains numeric data.

    Parameter
    ---------
    data : dask or numpy array

    Returns
    -------
    bool
        True or False

    """
    try:
        data.astype(float)
        return True
    except ValueError:
        return False


def _is_int(s):
    """Check that s in an integer.

    Parameter
    ---------
    s : python object to test

    Returns
    -------
    bool
        True or False

    """
    try:
        int(s)
        return True
    except ValueError:
        return False


def _check_search_keys(search_keys):
    if type(search_keys) is str:
        return [search_keys]
    elif type(search_keys) is list:
        if type(search_keys[0]) is not str:
            raise ValueError("key list provided is not a list of strings")
        else:
            return search_keys
    elif search_keys is None:
        return search_keys
    else:
        raise ValueError("search keys must be None, a string, "
                         "or a list of strings")


def _find_data(group, search_keys=None, hardlinks_only=False):
    """Read from a nexus or hdf file and return a list of the dataset entries.

    The method iterates through group attributes and returns NXdata or
    hdf datasets of size >=2 if they're not already NXdata blocks
    and returns a list of the entries
    This is a convenience method to inspect a file to see which datasets
    are present rather than loading all the sets in the file as signals
    h5py.visit or visititems does not visit soft
    links or external links so an implementation of a recursive
    search is required. See https://github.com/h5py/h5py/issues/671


    Parameters
    ----------
    group : hdf group or File
    search_keys  : string, list of strings or None, default: None
        Only return items which contain the strings
        .e.g search_list = ["instrument","Fe"] will return
        hdf entries with instrument or Fe in their hdf path.
    hardlinks_only : bool , default : False
        Option to ignore links (soft or External) within the file.

    Returns
    -------
    nx_dataset_list, hdf_dataset_list
        nx_dataset_list is a list of all NXdata paths
        hdf_dataset_list is a list of all hdf_datasets not linked to an
        NXdata set.

    """
    _check_search_keys(search_keys)
    all_hdf_datasets = []
    unique_hdf_datasets = []
    all_nx_datasets = []
    unique_nx_datasets = []
    rootname = ""

    def find_data_in_tree(group, rootname):
        for key, value in group.items():
            if rootname != "":
                rootkey = rootname + "/" + key
            else:
                rootkey = "/" + key
            if isinstance(value, h5py.Group):
                target = _getlink(group, rootkey, key)
                if "NX_class" in value.attrs:
                    if value.attrs["NX_class"] == b"NXdata" \
                            and "signal" in value.attrs.keys():
                        all_nx_datasets.append(rootkey)
                        if target is None:
                            unique_nx_datasets.append(rootkey)
                if hardlinks_only:
                    if target is None:
                        find_data_in_tree(value, rootkey)
                else:
                    find_data_in_tree(value, rootkey)
            else:
                if isinstance(value, h5py.Dataset):
                    if value.size >= 2:
                        target = _getlink(group, rootkey, key)
                        if not(value.dtype.type is np.string_ or
                                value.dtype.type is np.object_):
                            all_hdf_datasets.append(rootkey)
                            if target is None:
                                unique_hdf_datasets.append(rootkey)
    # need to use custom recursive function as visititems in h5py
    # does not visit links
    find_data_in_tree(group, rootname)

    if search_keys is None:
        # return all datasets
        if hardlinks_only:
            # return only the stored data, no linked data
            return unique_nx_datasets, unique_hdf_datasets
        else:
            return all_nx_datasets, all_hdf_datasets

    elif type(search_keys) is list:
        if hardlinks_only:
            # return only the stored data, no linked data
            nx_datasets = unique_nx_datasets
            hdf_datasets = unique_hdf_datasets
        else:
            nx_datasets = all_nx_datasets
            hdf_datasets = all_hdf_datasets
    # return data which contains a search string
    matched_hdf = [j for j in hdf_datasets
                   if any(s in j for s in search_keys)]
    matched_nexus = [j for j in nx_datasets
                     if any(s in j for s in search_keys)]
    return matched_nexus, matched_hdf


def _load_metadata(group, lazy=False):
    """Search through a hdf group and return the group structure.

    h5py.visit or visititems does not visit soft
    links or external links so an implementation of a recursive
    search is required. See https://github.com/h5py/h5py/issues/671

    Parameters
    ----------
    group : hdf group
        location to load the metadata from
    lazy : bool , default : False
        Option for lazy loading

    Returns
    -------
    dict
        dictionary of group contents


    """
    rootname = ""

    def find_meta_in_tree(group, rootname, lazy=False):
        tree = {}
        for key, item in group.attrs.items():
            new_key = _fix_exclusion_keys(key)
            if "attrs" not in tree.keys():
                tree["attrs"] = {}
            tree["attrs"][new_key] = _parse_from_file(item, lazy=lazy)

        for key, item in group.items():
            if rootname != "":
                rootkey = rootname + "/" + key
            else:
                rootkey = "/" + key
            new_key = _fix_exclusion_keys(key)
            if type(item) is h5py.Dataset:
                if item.attrs:
                    if new_key not in tree.keys():
                        tree[new_key] = {}
                    tree[new_key]["value"] = _parse_from_file(item, lazy=lazy)
                    for k, v in item.attrs.items():
                        if "attrs" not in tree[new_key].keys():
                            tree[new_key]["attrs"] = {}
                        tree[new_key]["attrs"][k] = _parse_from_file(v,
                                                                     lazy=lazy)
                else:
                    # this is to support hyperspy where datasets are not saved
                    # with attributes
                    tree[new_key] = _parse_from_file(item, lazy=lazy)
            elif type(item) is h5py.Group:
                if "NX_class" in item.attrs:
                    if item.attrs["NX_class"] != b"NXdata":
                        tree[new_key] = find_meta_in_tree(item, rootkey,
                                                          lazy=lazy)
                else:
                    tree[new_key] = find_meta_in_tree(item, rootkey,
                                                      lazy=lazy)

        return tree
    extracted_tree = find_meta_in_tree(group, rootname, lazy=lazy)
    return extracted_tree


def _fix_exclusion_keys(key):
    """Exclude hyperspy specific keys.

    Signal and DictionaryBrowser break if a
    a key is a dict method - e.g. {"keys":2.0}.

    This method prepends the key with "fix_" so the information is
    still present to work around this issue

    Parameter
    ---------
    key : str

    Returns
    -------
    str

    """
    if key.startswith("keys"):
        return "fix_"+key
    else:
        return key


def _find_search_keys_in_dict(tree, search_keys=None):
    """Search through a dict for search keys.

    This is a convenience method to inspect a file for a value
    rather than loading the file as a signal

    Parameters
    ----------
    tree         : h5py File object
    search_keys  : string or list of strings
        Only return items which contain the strings
        .e.g search_keys = ["instrument","Fe"] will return
        hdf entries with instrument or Fe in their hdf path.

    Returns
    -------
    dict
        When search_list is specified only full paths
        containing one or more search_keys will be returned

    """
    _check_search_keys(search_keys)
    metadata_dict = {}
    rootname = ""

    # recursive function
    def find_searchkeys_in_tree(myDict, rootname):
        for key, value in myDict.items():
            if rootname != "":
                rootkey = rootname + "/" + key
            else:
                rootkey = key
            if type(search_keys) is list \
                    and any([s1 in rootkey for s1 in search_keys]):
                mod_keys = _text_split(rootkey, (".", "/"))
                # create the key, values in the dict
                p = metadata_dict
                for d in mod_keys[:-1]:
                    p = p.setdefault(d, {})
                p[mod_keys[-1]] = value
            if isinstance(value, dict):
                find_searchkeys_in_tree(value, rootkey)

    if search_keys is None:
        return tree
    else:
        find_searchkeys_in_tree(tree, rootname)
    return metadata_dict


def _write_nexus_groups(dictionary, group, **kwds):
    """Recursively iterate throuh dictionary and write groups to nexus.

    Parameters
    ----------
    dictionary : dict
        dictionary contents to store to hdf group
    group : hdf group
        location to store dictionary
    **kwds : additional keywords
       additional keywords to pass to h5py.create_dataset method

    """
    for key, value in dictionary.items():
        if key == 'attrs':
            # we will handle attrs later...
            continue
        if isinstance(value, dict):
            if "attrs" in value:
                if "NX_class" in value["attrs"] and \
                        value["attrs"]["NX_class"] == "NXdata":
                    continue
            if 'value' in value.keys() \
                and not isinstance(value["value"], dict) \
                    and len(set(list(value.keys()) + ["attrs", "value"])) == 2:
                value = value["value"]
            else:
                _write_nexus_groups(value, group.require_group(key), **kwds)
        if isinstance(value, (list, tuple, np.ndarray, da.Array)):
            data = _parse_to_file(value)
            overwrite_dataset(group, data, key, chunks=None, **kwds)
        elif isinstance(value, (int, float, str, bytes)):
            group.create_dataset(key, data=_parse_to_file(value))


def _write_nexus_attr(dictionary, group):
    """Recursively iterate through dictionary and write "attrs" dictionaries.

    This step is called after the groups and datasets have been created

    Parameters
    ----------
    dictionary : dict
        Input dictionary to be written to the hdf group
    group : hdf group
        location to store the attrs sections of the dictionary

    """
    for key, value in dictionary.items():
        if key == 'attrs':

            for k, v in value.items():
                group.attrs[k] = _parse_to_file(v)
        else:
            if isinstance(value, dict):
                if "attrs" in value:
                    if "NX_class" in value["attrs"] and \
                            value["attrs"]["NX_class"] == "NXdata":
                        continue
                _write_nexus_attr(dictionary[key], group[key])


def read_metadata_from_file(filename, metadata_keys=None,
                            lazy=False, verbose=False):
    """Read the metadata from a nexus or hdf file.

    This method iterates through the file and returns a dictionary of
    the entries.
    This is a convenience method to inspect a file for a value
    rather than loading the file as a signal.

    Parameters
    ----------
    filename : str
        path of the file to read
    metadata_keys  : None,str or list_of_strings , default : None
        None will return all datasets found including linked data.
        Providing a string or list of strings will only return items
        which contain the string(s).
        For example, search_keys = ["instrument","Fe"] will return
        hdf entries with "instrument" or "Fe" in their hdf path.
    verbose: bool, default : False
        Pretty Print the results to screen

    Returns
    -------
    dict
        Metadata dictionary.

    See Also
    --------
    * :py:meth:`~.io_plugins.nexus.file_reader`
    * :py:meth:`~.io_plugins.nexus.file_writer`
    * :py:meth:`~.io_plugins.nexus.list_datasets_in_file`


    """
    search_keys = _check_search_keys(metadata_keys)
    fin = h5py.File(filename, "r")
    # search for NXdata sets...
    # strip out the metadata (basically everything other than NXdata)
    stripped_metadata = _load_metadata(fin, lazy=lazy)
    stripped_metadata = _find_search_keys_in_dict(stripped_metadata,
                                                  search_keys=search_keys)
    if verbose:
        pprint.pprint(stripped_metadata)
    return stripped_metadata


def list_datasets_in_file(filename, dataset_keys=None,
                          hardlinks_only=False,
                          verbose=True):
    """Read from a nexus or hdf file and return a list of the dataset paths.

    This method is used to inspect the contents of a Nexus file.
    The method iterates through group attributes and returns NXdata or
    hdf datasets of size >=2 if they're not already NXdata blocks
    and returns a list of the entries.
    This is a convenience method to inspect a file to list datasets
    present rather than loading all the datasets in the file as signals.

    Parameters
    ----------
    filename : str
        path of the file to read
    dataset_keys  : str, list of strings or None , default: None
        If a str or list of strings is provided only return items whose
        path contain the strings.
        For example, dataset_keys = ["instrument", "Fe"] will only return
        hdf entries with "instrument" or "Fe" somewhere in their hdf path.
    hardlinks_only : bool, default : False
        If true any links (soft or External) will be ignored when loading.
    verbose : boolean, default : True
        Prints the results to screen


    Returns
    -------
    list
        list of paths to datasets


    See Also
    --------
    * :py:meth:`~.io_plugins.nexus.file_reader`
    * :py:meth:`~.io_plugins.nexus.file_writer`
    * :py:meth:`~.io_plugins.nexus.read_metadata_from_file`


    """
    search_keys = _check_search_keys(dataset_keys)
    fin = h5py.File(filename, "r")
    # search for NXdata sets...
    # strip out the metadata (basically everything other than NXdata)
    nexus_data_paths, hdf_dataset_paths = \
        _find_data(fin, search_keys=search_keys, hardlinks_only=hardlinks_only)
    if verbose:
        if nexus_data_paths:
            print("NXdata found")
            for nxd in nexus_data_paths:
                print(nxd)
        else:
            print("NXdata not found")
        if hdf_dataset_paths:
            print("Unique HDF datasets found")
            for hdfd in hdf_dataset_paths:
                print(hdfd, fin[hdfd].shape)
        else:
            print("No HDF datasets not found or data is captured by NXdata")
    fin.close()
    return nexus_data_paths, hdf_dataset_paths


def _write_signal(signal, nxgroup, signal_name, **kwds):
    """Store the signal data as an NXdata dataset.

    Parameters
    ----------
    signal : Hyperspy signal
    nxgroup : HDF group
        Entry at which to save signal data
    signal_name : str
        Name  under which to store the signal entry in the file

    """
    smd = signal.metadata.Signal
    if signal.axes_manager.signal_dimension == 1:
        smd.record_by = "spectrum"
    elif signal.axes_manager.signal_dimension == 2:
        smd.record_by = "image"
    else:
        smd.record_by = ""

    nxdata = nxgroup.require_group(signal_name)
    nxdata.attrs["NX_class"] = _parse_to_file("NXdata")
    nxdata.attrs["signal"] = _parse_to_file("data")
    if smd.record_by:
        nxdata.attrs["interpretation"] = _parse_to_file(smd.record_by)
    overwrite_dataset(nxdata, signal.data, "data", chunks=None, **kwds)
    axis_names = [_parse_to_file(".")] * len(signal.axes_manager.shape)
    for i, axis in enumerate(signal.axes_manager._axes):
        if axis.name != t.Undefined:
            axname = axis.name
            axindex = [axis.index_in_array]
            indices = _parse_to_file(axis.name + "_indices")
            nxdata.attrs[indices] = _parse_to_file(axindex)
            nxdata.require_dataset(axname, data=axis.axis,
                                   shape=axis.axis.shape,
                                   dtype=axis.axis.dtype)
            axis_names[axis.index_in_array] = axname

    nxdata.attrs["axes"] = _parse_to_file(axis_names)
    return nxdata


def file_writer(filename,
                signals,
                save_original_metadata=True,
                use_default=False,
                *args, **kwds):
    """Write the signal and metadata as a nexus file.

    This will save the signal in NXdata format in the file.
    As the form of the metadata can vary and is not validated it will
    be stored as an NXcollection (an unvalidated collection)

    Parameters
    ----------
    filename : str
        Path of the file to write
    signals : signal or list of signals
        Signal(s) to be written
    save_original_metadata : bool , default : False
          Option to save hyperspy.original_metadata with the signal.
          A loaded Nexus file may have a large amount of data
          when loaded which you may wish to omit on saving
    use_default : bool , default : False
          Option to define the default dataset in the file.
          If set to True the signal or first signal in the list of signals
          will be defined as the default (following Nexus v3 data rules).

    See Also
    --------
    * :py:meth:`~.io_plugins.nexus.file_reader`
    * :py:meth:`~.io_plugins.nexus.list_datasets_in_file`
    * :py:meth:`~.io_plugins.nexus.read_metadata_from_file`

    """
    if not isinstance(signals, list):
        signals = [signals]

    with h5py.File(filename, mode='w') as f:
        f.attrs['file_format'] = "nexus"
        f.attrs['file_writer'] = "hyperspy_nexus_v3"
        if 'compression' not in kwds:
            kwds['compression'] = 'gzip'
        if use_default:
            f.attrs["default"] = "entry1"
        #
        # write the signals
        #

        for i, sig in enumerate(signals):
            nxentry = f.create_group("entry%d" % (i + 1))
            nxentry.attrs["NX_class"] = _parse_to_file("NXentry")
            signal_name = sig.metadata.General.title \
                if sig.metadata.General.title else 'unnamed__%d' % i
            if "/" in signal_name:
                signal_name = signal_name.replace("/", "_")
            if signal_name.startswith("__"):
                signal_name = signal_name[2:]

            if i == 0 and use_default:
                nxentry.attrs["default"] = signal_name

            nxaux = nxentry.create_group("auxiliary")
            nxaux.attrs["NX_class"] = _parse_to_file("NXentry")
            _write_signal(sig, nxentry, signal_name, **kwds)

            if sig.learning_results:
                nxlearn = nxaux.create_group('learning_results')
                nxlearn.attrs["NX_class"] = _parse_to_file("NXcollection")
                learn = sig.learning_results.__dict__
                _write_nexus_groups(learn, nxlearn, **kwds)
                _write_nexus_attr(learn, nxlearn)
            #
            # write metadata
            #
            if save_original_metadata:
                if sig.original_metadata:
                    if isinstance(sig.original_metadata,
                                  DictionaryTreeBrowser):
                        ometa = sig.original_metadata.as_dictionary()
                    else:
                        ometa = sig.original_metadata

                    nxometa = nxaux.create_group('original_metadata')
                    nxometa.attrs["NX_class"] = _parse_to_file("NXcollection")
                    # write the groups and structure
                    _write_nexus_groups(ometa, nxometa, **kwds)
                    _write_nexus_attr(ometa, nxometa)

            if sig.metadata:
                if isinstance(sig.metadata, DictionaryTreeBrowser):
                    meta = sig.metadata.as_dictionary()
                else:
                    meta = sig.metadata

                nxometa = nxaux.create_group('hyperspy_metadata')
                nxometa.attrs["NX_class"] = _parse_to_file("NXcollection")
                # write the groups and structure
                _write_nexus_groups(meta, nxometa, **kwds)
                _write_nexus_attr(meta, nxometa)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging
from collections import namedtuple

import numpy as np
import pytest

import hyperspy.api as hs
from hyperspy import _lazy_signals
from hyperspy.exceptions import VisibleDeprecationWarning
from hyperspy.io import assign_signal_subclass

testcase = namedtuple('testcase', ['dtype', 'sig_dim', 'sig_type', 'cls'])

subclass_cases = (
    testcase("float", 1000, "", "BaseSignal"),
    testcase("float", 1, "", "Signal1D"),
    testcase("float", 2, "", "Signal2D"),
    testcase("float", 1, "EELS", "EELSSpectrum"),
    testcase("float", 1, "EDS_SEM", "EDSSEMSpectrum"),
    testcase("float", 1, "EDS_TEM", "EDSTEMSpectrum"),
    testcase("complex", 1, "DielectricFunction", "DielectricFunction"),
    testcase("complex", 1, "dielectric function", "DielectricFunction"),
    testcase("complex", 1000, "", "ComplexSignal"),
    testcase("complex", 1, "", "ComplexSignal1D"),
    testcase("complex", 2, "", "ComplexSignal2D"),
    testcase("float", 1000, "DefinitelyNotAHyperSpySignal", "BaseSignal"),
    testcase("float", 1, "DefinitelyNotAHyperSpySignal", "Signal1D"),
    testcase("complex", 1000, "DefinitelyNotAHyperSpySignal", "ComplexSignal"),
    testcase("float", 1, "DefinitelyNotAHyperSpySignal", "Signal1D"),
)

def test_assignment_class(caplog):
    for case in subclass_cases:
        with caplog.at_level(logging.WARNING):
            new_subclass = assign_signal_subclass(
                 dtype=np.dtype(case.dtype),
                 signal_dimension=case.sig_dim,
                 signal_type=case.sig_type,
                 lazy=False,
            )

        assert new_subclass is getattr(hs.signals, case.cls)

        warn_msg = "not understood. See `hs.print_known_signal_types()` for a list of known signal types"
        if case.sig_type == "DefinitelyNotAHyperSpySignal":
            assert warn_msg in caplog.text
        else:
            assert warn_msg not in caplog.text

        lazyclass = 'Lazy' + case.cls if case.cls != 'BaseSignal' else 'LazySignal'
        new_subclass = assign_signal_subclass(
            dtype=np.dtype(case.dtype),
            signal_dimension=case.sig_dim,
            signal_type=case.sig_type,
            lazy=True,
        )

        assert new_subclass is getattr(_lazy_signals, lazyclass)


class TestConvertBaseSignal:

    def setup_method(self, method):
        self.s = hs.signals.BaseSignal(np.zeros((3, 3)))

    def test_base_to_lazy(self):
        assert not self.s._lazy
        self.s._lazy = True
        self.s._assign_subclass()
        assert isinstance(self.s, _lazy_signals.LazySignal)
        assert self.s._lazy

    def test_base_to_1d(self):
        self.s.axes_manager.set_signal_dimension(1)
        self.s._assign_subclass()
        assert isinstance(self.s, hs.signals.Signal1D)
        self.s.metadata.Signal.record_by = ''
        self.s._assign_subclass()
        assert isinstance(self.s, hs.signals.BaseSignal)

    def test_base_to_2d(self):
        self.s.axes_manager.set_signal_dimension(2)
        self.s._assign_subclass()
        assert isinstance(self.s, hs.signals.Signal2D)

    def test_base_to_complex(self):
        self.s.change_dtype(complex)
        assert isinstance(self.s, hs.signals.ComplexSignal)
        # Going back from ComplexSignal to BaseSignal is not possible!
        # If real data is required use `real`, `imag`, `amplitude` or `phase`!


class TestConvertSignal1D:

    def setup_method(self, method):
        self.s = hs.signals.Signal1D([0])

    def test_lazy_to_eels_and_back(self):
        self.s = self.s.as_lazy()
        self.s.set_signal_type("EELS")
        assert isinstance(self.s, _lazy_signals.LazyEELSSpectrum)
        self.s.set_signal_type("")
        assert isinstance(self.s, _lazy_signals.LazySignal1D)

    def test_signal1d_to_eels(self):
        self.s.set_signal_type("EELS")
        assert isinstance(self.s, hs.signals.EELSSpectrum)
        self.s.set_signal_type("")
        assert isinstance(self.s, hs.signals.Signal1D)

    def test_signal1d_to_eds_tem(self):
        self.s.set_signal_type("EDS_TEM")
        assert isinstance(self.s, hs.signals.EDSTEMSpectrum)
        self.s.set_signal_type("")
        assert isinstance(self.s, hs.signals.Signal1D)

    def test_signal1d_to_eds_sem(self):
        self.s.set_signal_type("EDS_SEM")
        assert isinstance(self.s, hs.signals.EDSSEMSpectrum)
        self.s.set_signal_type("")
        assert isinstance(self.s, hs.signals.Signal1D)

    def test_deprecated(self):
        with pytest.warns(
            VisibleDeprecationWarning,
            match=r"is deprecated. Use ",
        ):
            self.s.set_signal_type(None)


class TestConvertComplexSignal:

    def setup_method(self, method):
        self.s = hs.signals.ComplexSignal(np.zeros((3, 3)))

    def test_complex_to_complex1d(self):
        self.s.axes_manager.set_signal_dimension(1)
        self.s._assign_subclass()
        assert isinstance(self.s, hs.signals.ComplexSignal1D)

    def test_complex_to_complex2d(self):
        self.s.axes_manager.set_signal_dimension(2)
        self.s._assign_subclass()
        assert isinstance(self.s, hs.signals.ComplexSignal2D)


class TestConvertComplexSignal1D:

    def setup_method(self, method):
        self.s = hs.signals.ComplexSignal1D([0])

    def test_complex_to_dielectric_function(self):
        self.s.set_signal_type("DielectricFunction")
        assert isinstance(self.s, hs.signals.DielectricFunction)
        self.s.set_signal_type("")
        assert isinstance(self.s, hs.signals.ComplexSignal1D)



# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import os
from shutil import copyfile
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
import scipy.misc

import hyperspy.api as hs
from hyperspy.misc.test_utils import update_close_figure
from hyperspy.signals import Signal1D
from hyperspy.drawing.signal1d import Signal1DLine
from hyperspy.tests.drawing.test_plot_signal import _TestPlot

scalebar_color = 'blue'
default_tol = 2.0
baseline_dir = 'plot_signal1d'
style_pytest_mpl = 'default'

style = ['default', 'overlap', 'cascade', 'mosaic', 'heatmap']


@pytest.fixture
def mpl_generate_path_cmdopt(request):
    return request.config.getoption("--mpl-generate-path")


def _generate_filename_list(style):
    path = Path(__file__).resolve().parent
    baseline_path = path.joinpath(baseline_dir)

    filename_list = [f'test_plot_spectra_{s}' for s in style] + \
                    [f'test_plot_spectra_rev_{s}' for s in style]
    filename_list2 = []

    for filename in filename_list:
        for i in range(0, 4):
            filename_list2.append(
                baseline_path.joinpath(f'{filename}{i}.png')
            )

    return filename_list2


@pytest.fixture
def setup_teardown(request, scope="class"):
    mpl_generate_path_cmdopt = request.config.getoption("--mpl-generate-path")
    # SETUP
    # duplicate baseline images to match the test_name when the
    # parametrized 'test_plot_spectra' are run. For a same 'style', the
    # expected images are the same.
    if mpl_generate_path_cmdopt is None:
        for filename in _generate_filename_list(style):
            copyfile(f"{str(filename)[:-5]}.png", filename)
    yield
    # TEARDOWN
    # Create the baseline images: copy one baseline image for each test
    # and remove the other ones.
    if mpl_generate_path_cmdopt:
        for filename in _generate_filename_list(style):
            copyfile(filename, f"{str(filename)[:-5]}.png")
    # Delete the images that have been created in 'setup_class'
    for filename in _generate_filename_list(style):
        os.remove(filename)


@pytest.mark.usefixtures("setup_teardown")
class TestPlotSpectra():

    s = hs.signals.Signal1D(scipy.misc.ascent()[100:160:10])

    # Add a test signal with decreasing axis
    s_reverse = s.deepcopy()
    s_reverse.axes_manager[1].offset = 512
    s_reverse.axes_manager[1].scale = -1

    def _generate_parameters(style):
        parameters = []
        for s in style:
            for fig in [True, None]:
                for ax in [True, None]:
                    parameters.append([s, fig, ax])
        return parameters

    def _generate_ids(style, duplicate=4):
        ids = []
        for s in style:
            ids.extend([s] * duplicate)
        return ids

    @pytest.mark.parametrize(("style", "fig", "ax"),
                             _generate_parameters(style),
                             ids=_generate_ids(style))
    @pytest.mark.mpl_image_compare(baseline_dir=baseline_dir,
                                   tolerance=default_tol, style=style_pytest_mpl)
    def test_plot_spectra(self, style, fig, ax):
        if fig:
            fig = plt.figure()
        if ax:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        ax = hs.plot.plot_spectra(self.s, style=style, legend='auto',
                                  fig=fig, ax=ax)
        if style == 'mosaic':
            ax = ax[0]
        return ax.figure

    @pytest.mark.parametrize(("style", "fig", "ax"),
                             _generate_parameters(style),
                             ids=_generate_ids(style))
    @pytest.mark.mpl_image_compare(baseline_dir=baseline_dir,
                                   tolerance=default_tol, style=style_pytest_mpl)
    def test_plot_spectra_rev(self, style, fig, ax):
        if fig:
            fig = plt.figure()
        if ax:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        ax = hs.plot.plot_spectra(self.s_reverse, style=style, legend='auto',
                                  fig=fig, ax=ax)
        if style == 'mosaic':
            ax = ax[0]
        return ax.figure

    @pytest.mark.parametrize("figure", ['1nav', '1sig', '2nav', '2sig'])
    @pytest.mark.mpl_image_compare(baseline_dir=baseline_dir,
                                   tolerance=default_tol, style=style_pytest_mpl)
    def test_plot_spectra_sync(self, figure):
        s1 = hs.signals.Signal1D(scipy.misc.face()).as_signal1D(0).inav[:, :3]
        s2 = s1.deepcopy() * -1
        hs.plot.plot_signals([s1, s2])
        if figure == '1nav':
            return s1._plot.signal_plot.figure
        if figure == '1sig':
            return s1._plot.navigator_plot.figure
        if figure == '2nav':
            return s2._plot.navigator_plot.figure
        if figure == '2sig':
            return s2._plot.navigator_plot.figure

    def test_plot_spectra_legend_pick(self):
        x = np.linspace(0., 2., 512)
        n = np.arange(1, 5)
        x_pow_n = x[None, :]**n[:, None]
        s = hs.signals.Signal1D(x_pow_n)
        my_legend = [r'x^' + str(io) for io in n]
        f = plt.figure()
        ax = hs.plot.plot_spectra(s, legend=my_legend, fig=f)
        leg = ax.get_legend()
        leg_artists = leg.get_lines()
        click = plt.matplotlib.backend_bases.MouseEvent(
            'button_press_event', f.canvas, 0, 0, 'left')
        for artist, li in zip(leg_artists, ax.lines[::-1]):
            plt.matplotlib.backends.backend_agg.FigureCanvasBase.pick_event(
                f.canvas, click, artist)
            assert not li.get_visible()
            plt.matplotlib.backends.backend_agg.FigureCanvasBase.pick_event(
                f.canvas, click, artist)

    @pytest.mark.mpl_image_compare(baseline_dir=baseline_dir,
                                   tolerance=default_tol, style=style_pytest_mpl)
    def test_plot_spectra_auto_update(self):
        s = hs.signals.Signal1D(np.arange(100))
        s2 = s / 2
        ax = hs.plot.plot_spectra([s, s2])
        s.data = -s.data
        s.events.data_changed.trigger(s)
        s2.data = -s2.data * 4 + 50
        s2.events.data_changed.trigger(s2)

        return ax.get_figure()


@update_close_figure
def test_plot_nav0_close():
    test_plot = _TestPlot(ndim=0, sdim=1)
    test_plot.signal.plot()
    return test_plot.signal


@update_close_figure
def test_plot_nav1_close():
    test_plot = _TestPlot(ndim=1, sdim=1)
    test_plot.signal.plot()
    return test_plot.signal


@update_close_figure
def test_plot_nav2_close():
    test_plot = _TestPlot(ndim=2, sdim=1)
    test_plot.signal.plot()
    return test_plot.signal


def _test_plot_two_cursors(ndim):
    test_plot = _TestPlot(ndim=ndim, sdim=1)  # sdim=2 not supported
    s = test_plot.signal
    s.metadata.General.title = 'Nav %i, Sig 1, two cursor' % ndim
    s.axes_manager[0].index = 4
    s.plot()
    s._plot.add_right_pointer()
    s._plot.navigator_plot.figure.canvas.draw()
    s._plot.signal_plot.figure.canvas.draw()
    s._plot.right_pointer.axes_manager[0].index = 2
    if ndim == 2:
        s.axes_manager[1].index = 2
        s._plot.right_pointer.axes_manager[1].index = 3
    return s


@pytest.mark.parametrize('autoscale', ['', 'x', 'xv', 'v'])
@pytest.mark.parametrize('norm', ['log', 'auto'])
def test_plot_two_cursos_parameters(autoscale, norm):
    kwargs = {'autoscale':autoscale, 'norm':norm}
    test_plot = _TestPlot(ndim=2, sdim=1)  # sdim=2 not supported
    s = test_plot.signal
    s.plot(**kwargs)
    s._plot.add_right_pointer(**kwargs)
    for line in s._plot.signal_plot.ax_lines:
        assert line.autoscale == autoscale


def _generate_parameter():
    parameters = []
    for ndim in [1, 2]:
        for plot_type in ['nav', 'sig']:
            parameters.append([ndim, plot_type])
    return parameters


@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir,
                               tolerance=default_tol, style=style_pytest_mpl)
def test_plot_log_scale():
    s = Signal1D(np.exp(-np.arange(100) / 5.0))
    s.plot(norm='log')
    return s._plot.signal_plot.figure


@pytest.mark.parametrize(("ndim", "plot_type"),
                         _generate_parameter())
@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir,
                               tolerance=default_tol, style=style_pytest_mpl)
def test_plot_two_cursors(ndim, plot_type):
    s = _test_plot_two_cursors(ndim=ndim)

    if plot_type == "sig":
        f = s._plot.signal_plot.figure
    else:
        f= s._plot.navigator_plot.figure
    f.canvas.draw()
    f.canvas.flush_events()
    return f


@update_close_figure
def test_plot_nav2_sig1_two_cursors_close():
    return _test_plot_two_cursors(ndim=2)


def test_plot_with_non_finite_value():
    s = hs.signals.Signal1D(np.array([np.nan, 2.0]))
    s.plot()
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)

    s = hs.signals.Signal1D(np.array([np.nan, np.nan]))
    s.plot()
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)

    s = hs.signals.Signal1D(np.array([-np.inf, 2.0]))
    s.plot()
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)

    s = hs.signals.Signal1D(np.array([np.inf, 2.0]))
    s.plot()
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)


def test_plot_add_line_events():
    s = hs.signals.Signal1D(np.arange(100))
    s.plot()
    assert len(s.axes_manager.events.indices_changed.connected) == 1
    figure = s._plot.signal_plot

    def line_function(axes_manager=None):
        return 100 - np.arange(100)

    line = Signal1DLine()
    line.data_function = line_function
    line.set_line_properties(color='blue', type='line', scaley=False)
    figure.add_line(line, connect_navigation=True)
    line.plot()
    assert len(line.events.closed.connected) == 1
    assert len(s.axes_manager.events.indices_changed.connected) == 2

    line.close()
    figure.close_right_axis()

    assert len(line.events.closed.connected) == 0
    assert len(s.axes_manager.events.indices_changed.connected) == 1

    figure.close()
    assert len(s.axes_manager.events.indices_changed.connected) == 0


@pytest.mark.parametrize("autoscale", ['', 'x', 'xv', 'v'])
@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir,
                               tolerance=default_tol, style=style_pytest_mpl)
def test_plot_autoscale(autoscale):
    s = hs.datasets.artificial_data.get_core_loss_eels_line_scan_signal(
        add_powerlaw=True, add_noise=False)
    s.plot(autoscale=autoscale)
    ax = s._plot.signal_plot.ax
    ax.set_xlim(500.0, 700.0)
    ax.set_ylim(-10.0, 20.0)
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)

    return s._plot.signal_plot.figure



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from hyperspy.exceptions import VisibleDeprecationWarning
from hyperspy.misc.machine_learning.import_sklearn import sklearn_installed
from hyperspy.signals import Signal1D


class TestLazyDecomposition:
    def setup_method(self, method):
        # Define shape etc.
        m = 100  # Dimensionality
        n = 128  # Number of samples
        r = 3

        self.rng = np.random.RandomState(101)
        U = self.rng.randn(m, r)
        V = self.rng.randn(n, r)
        X = U @ V.T
        X = np.exp(0.1 * X / np.linalg.norm(X))

        self.m = m
        self.n = n
        self.rank = r
        self.X = X
        self.s = Signal1D(
            X.copy().reshape(int(np.sqrt(m)), int(np.sqrt(m)), n)
        ).as_lazy()

        # Test tolerance
        self.tol = 1e-2 * (self.m * self.n)

    @pytest.mark.parametrize("output_dimension", [None, 3])
    @pytest.mark.parametrize("normalize_poissonian_noise", [True, False])
    def test_svd(self, output_dimension, normalize_poissonian_noise):
        self.s.decomposition(
            output_dimension=output_dimension,
            normalize_poissonian_noise=normalize_poissonian_noise,
        )
        factors = self.s.learning_results.factors
        loadings = self.s.learning_results.loadings

        if hasattr(factors, "compute"):
            factors = factors.compute()
        if hasattr(loadings, "compute"):
            loadings = loadings.compute()

        explained_variance = self.s.learning_results.explained_variance
        X = loadings @ factors.T

        # Check the low-rank component MSE
        normX = np.linalg.norm(X - self.X)
        assert normX < self.tol

        # Check singular values
        explained_variance_norm = explained_variance / np.sum(explained_variance)
        np.testing.assert_allclose(
            explained_variance_norm[: self.rank].sum(), 1.0, atol=1e-6
        )

    @pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
    @pytest.mark.parametrize("normalize_poissonian_noise", [True, False])
    def test_pca(self, normalize_poissonian_noise):
        self.s.decomposition(
            output_dimension=3,
            algorithm="PCA",
            normalize_poissonian_noise=normalize_poissonian_noise,
        )
        factors = self.s.learning_results.factors
        loadings = self.s.learning_results.loadings

        if hasattr(factors, "compute"):
            factors = factors.compute()
        if hasattr(loadings, "compute"):
            loadings = loadings.compute()

        explained_variance = self.s.learning_results.explained_variance
        X = loadings @ factors.T

        # Check the low-rank component MSE
        normX = np.linalg.norm(X - self.X)
        assert normX < self.tol

        # Check singular values
        explained_variance_norm = explained_variance / np.sum(explained_variance)
        np.testing.assert_allclose(
            explained_variance_norm[: self.rank].sum(), 1.0, atol=1e-6
        )

    @pytest.mark.parametrize("normalize_poissonian_noise", [True, False])
    def test_orpca(self, normalize_poissonian_noise):
        self.s.decomposition(
            output_dimension=3,
            algorithm="ORPCA",
            normalize_poissonian_noise=normalize_poissonian_noise,
        )
        factors = self.s.learning_results.factors
        loadings = self.s.learning_results.loadings

        if hasattr(factors, "compute"):
            factors = factors.compute()
        if hasattr(loadings, "compute"):
            loadings = loadings.compute()

        explained_variance = self.s.learning_results.explained_variance
        X = loadings @ factors.T

        # Check the low-rank component MSE
        normX = np.linalg.norm(X - self.X)
        assert normX < self.tol

        # Check singular values
        assert explained_variance is None

    @pytest.mark.parametrize("normalize_poissonian_noise", [True, False])
    def test_ornmf(self, normalize_poissonian_noise):
        self.s.decomposition(
            output_dimension=3,
            algorithm="ORNMF",
            normalize_poissonian_noise=normalize_poissonian_noise,
        )
        factors = self.s.learning_results.factors
        loadings = self.s.learning_results.loadings

        if hasattr(factors, "compute"):
            factors = factors.compute()
        if hasattr(loadings, "compute"):
            loadings = loadings.compute()

        explained_variance = self.s.learning_results.explained_variance
        X = loadings @ factors.T

        # Check the low-rank component MSE
        normX = np.linalg.norm(X - self.X)
        assert normX < self.tol

        # Check singular values
        assert explained_variance is None

    def test_output_dimension_error(self):
        with pytest.raises(ValueError, match="`output_dimension` must be specified"):
            self.s.decomposition(algorithm="ORPCA")

    def test_algorithm_error(self):
        with pytest.raises(ValueError, match="'algorithm' not recognised"):
            self.s.decomposition(algorithm="random")

    def test_bounds_warning(self):
        with pytest.warns(
            VisibleDeprecationWarning, match="`bounds` keyword is deprecated"
        ):
            self.s.decomposition(bounds=True)

    @pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
    @pytest.mark.parametrize("algorithm", ["ONMF"])
    def test_deprecated_algorithms_warning(self, algorithm):
        with pytest.warns(
            VisibleDeprecationWarning,
            match="`algorithm='{}'` has been deprecated".format(algorithm),
        ):
            self.s.decomposition(output_dimension=3, algorithm=algorithm)


class TestPrintInfo:
    def setup_method(self, method):
        rng = np.random.RandomState(123)
        self.s = Signal1D(rng.random_sample(size=(20, 100))).as_lazy()

    @pytest.mark.parametrize("algorithm", ["SVD", "ORPCA", "ORNMF"])
    def test_decomposition(self, algorithm, capfd):
        self.s.decomposition(algorithm=algorithm, output_dimension=3)
        captured = capfd.readouterr()
        assert "Decomposition info:" in captured.out

    @pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
    @pytest.mark.parametrize("algorithm", ["PCA"])
    def test_decomposition_sklearn(self, capfd, algorithm):
        self.s.decomposition(algorithm=algorithm, output_dimension=3)
        captured = capfd.readouterr()
        assert "Decomposition info:" in captured.out
        assert "scikit-learn estimator:" in captured.out

    @pytest.mark.parametrize("algorithm", ["SVD"])
    def test_no_print(self, algorithm, capfd):
        self.s.decomposition(algorithm=algorithm, output_dimension=2, print_info=False)
        captured = capfd.readouterr()
        assert "Decomposition info:" not in captured.out
